@@ -6,38 +6,53 @@
  */
 #include "waylandimserver.h"
 #include <sys/mman.h>
+#include <memory>
 #include <fcitx-utils/utf8.h>
+#include "appmonitor.h"
+#include "virtualinputcontext.h"
+#include "wayland-text-input-unstable-v1-client-protocol.h"
 #include "waylandim.h"
 
+#ifdef __linux__
+#include <linux/input-event-codes.h>
+#elif __FreeBSD__
+#include <dev/evdev/input-event-codes.h>
+#else
+#define BTN_LEFT 0x110
+#define BTN_RIGHT 0x111
+#endif
+
 namespace fcitx {
 
+constexpr CapabilityFlags baseFlags{CapabilityFlag::Preedit,
+                                    CapabilityFlag::FormattedPreedit,
+                                    CapabilityFlag::SurroundingText};
+
 static inline unsigned int waylandFormat(TextFormatFlags flags) {
-    unsigned int result = 0;
-    if (flags & TextFormatFlag::Underline) {
-        result |= ZWP_TEXT_INPUT_V1_PREEDIT_STYLE_UNDERLINE;
-    }
     if (flags & TextFormatFlag::HighLight) {
-        result |= ZWP_TEXT_INPUT_V1_PREEDIT_STYLE_SELECTION;
+        return ZWP_TEXT_INPUT_V1_PREEDIT_STYLE_HIGHLIGHT;
     }
     if (flags & TextFormatFlag::Bold) {
-        result |= ZWP_TEXT_INPUT_V1_PREEDIT_STYLE_ACTIVE;
+        return ZWP_TEXT_INPUT_V1_PREEDIT_STYLE_ACTIVE;
     }
     if (flags & TextFormatFlag::Strike) {
-        result |= ZWP_TEXT_INPUT_V1_PREEDIT_STYLE_INCORRECT;
+        return ZWP_TEXT_INPUT_V1_PREEDIT_STYLE_INCORRECT;
     }
-    return result;
+    if (flags & TextFormatFlag::Underline) {
+        return ZWP_TEXT_INPUT_V1_PREEDIT_STYLE_DEFAULT;
+    }
+    return ZWP_TEXT_INPUT_V1_PREEDIT_STYLE_NONE;
 }
 
 WaylandIMServer::WaylandIMServer(wl_display *display, FocusGroup *group,
                                  const std::string &name,
                                  WaylandIMModule *waylandim)
-    : group_(group), name_(name), parent_(waylandim), inputMethodV1_(nullptr),
-      display_(
-          static_cast<wayland::Display *>(wl_display_get_user_data(display))) {
+    : WaylandIMServerBase(display, group, name, waylandim),
+      inputMethodV1_(nullptr) {
     display_->requestGlobals<wayland::ZwpInputMethodV1>();
-    globalConn_ = display_->registry()->global().connect(
-        [this](uint32_t, const char *interface, uint32_t) {
-            if (0 == strcmp(interface, wayland::ZwpInputMethodV1::interface)) {
+    globalConn_ = display_->globalCreated().connect(
+        [this](const std::string &interface, const std::shared_ptr<void> &) {
+            if (interface == wayland::ZwpInputMethodV1::interface) {
                 init();
             }
         });
@@ -45,7 +60,8 @@ WaylandIMServer::WaylandIMServer(wl_display *display, FocusGroup *group,
     init();
 }
 
-WaylandIMServer::~WaylandIMServer() { delete globalIc_; }
+WaylandIMServer::~WaylandIMServer() { delete globalIc_.get(); }
+
 InputContextManager &WaylandIMServer::inputContextManager() {
     return parent_->instance()->inputContextManager();
 }
@@ -55,37 +71,58 @@ Instance *WaylandIMServer::instance() { return parent_->instance(); }
 void WaylandIMServer::init() {
     auto im = display_->getGlobal<wayland::ZwpInputMethodV1>();
     if (im && !inputMethodV1_) {
+        WAYLANDIM_DEBUG() << "WAYLANDIM V1";
         inputMethodV1_ = im;
-        globalIc_ = new WaylandIMInputContextV1(
+        auto *globalIc = new WaylandIMInputContextV1(
             parent_->instance()->inputContextManager(), this);
-        globalIc_->setFocusGroup(group_);
+        globalIc->setFocusGroup(group_);
+        globalIc->setCapabilityFlags(baseFlags);
+        globalIc_ = globalIc->watch();
         inputMethodV1_->activate().connect(
-            [this](wayland::ZwpInputMethodContextV1 *ic) { activate(ic); });
+            [this](wayland::ZwpInputMethodContextV1 *ic) {
+                WAYLANDIM_DEBUG() << "ACTIVATE " << ic;
+                activate(ic);
+            });
         inputMethodV1_->deactivate().connect(
-            [this](wayland::ZwpInputMethodContextV1 *ic) { deactivate(ic); });
-        display_->flush();
+            [this](wayland::ZwpInputMethodContextV1 *ic) {
+                WAYLANDIM_DEBUG() << "DEACTIVATE " << ic;
+                deactivate(ic);
+            });
+        deferredFlush();
     }
 }
 
 void WaylandIMServer::activate(wayland::ZwpInputMethodContextV1 *id) {
-    globalIc_->activate(id);
+    if (auto *globalIc =
+            static_cast<WaylandIMInputContextV1 *>(globalIc_.get())) {
+        globalIc->activate(id);
+    }
 }
 
 void WaylandIMServer::deactivate(wayland::ZwpInputMethodContextV1 *id) {
-    globalIc_->deactivate(id);
+    if (auto *globalIc =
+            static_cast<WaylandIMInputContextV1 *>(globalIc_.get())) {
+        globalIc->deactivate(id);
+    }
 }
 
 WaylandIMInputContextV1::WaylandIMInputContextV1(
     InputContextManager &inputContextManager, WaylandIMServer *server)
-    : InputContext(inputContextManager), server_(server) {
+    : VirtualInputContextGlue(inputContextManager), server_(server) {
     timeEvent_ = server_->instance()->eventLoop().addTimeEvent(
         CLOCK_MONOTONIC, now(CLOCK_MONOTONIC), 0,
         [this](EventSourceTime *, uint64_t) {
             repeat();
             return true;
         });
+    timeEvent_->setAccuracy(1);
     timeEvent_->setEnabled(false);
     created();
+
+    if (auto *appMonitor = server->parent_->appMonitor(server->name_)) {
+        virtualICManager_ = std::make_unique<VirtualInputContextManager>(
+            &inputContextManager, this, appMonitor);
+    }
 }
 
 WaylandIMInputContextV1::~WaylandIMInputContextV1() { destroy(); }
@@ -100,13 +137,14 @@ void WaylandIMInputContextV1::activate(wayland::ZwpInputMethodContextV1 *ic) {
     ic_->contentType().connect([this](uint32_t hint, uint32_t purpose) {
         contentTypeCallback(hint, purpose);
     });
-    ic_->invokeAction().connect([](uint32_t button, uint32_t index) {
+    ic_->invokeAction().connect([this](uint32_t button, uint32_t index) {
         invokeActionCallback(button, index);
     });
     ic_->commitState().connect(
         [this](uint32_t serial) { commitStateCallback(serial); });
     ic_->preferredLanguage().connect(
         [](const char *language) { preferredLanguageCallback(language); });
+    keyboard_.reset();
     keyboard_.reset(ic_->grabKeyboard());
     keyboard_->keymap().connect(
         [this](uint32_t format, int32_t fd, uint32_t size) {
@@ -126,15 +164,33 @@ void WaylandIMInputContextV1::activate(wayland::ZwpInputMethodContextV1 *ic) {
         repeatInfoCallback(rate, delay);
     });
     repeatInfoCallback(repeatRate_, repeatDelay_);
-    server_->display_->sync();
-    focusIn();
+    wl_array array;
+    wl_array_init(&array);
+    constexpr char data[] = "Shift\0Control\0Mod1\0Mod4";
+    wl_array_add(&array, sizeof(data));
+    memcpy(array.data, data, sizeof(data));
+    ic_->modifiersMap(&array);
+    wl_array_release(&array);
+    if (virtualICManager_) {
+        virtualICManager_->setRealFocus(true);
+    } else {
+        focusIn();
+    }
+    server_->deferredFlush();
 }
 
 void WaylandIMInputContextV1::deactivate(wayland::ZwpInputMethodContextV1 *ic) {
     if (ic_.get() == ic) {
         ic_.reset();
         keyboard_.reset();
-        focusOut();
+        // This is the only place we update wayland display mask, so it is ok to
+        // reset it to 0. This breaks the caps lock or num lock. But we have no
+        // other option until we can listen to the mod change globally.
+        server_->instance()->clearXkbStateMask(server_->group()->display());
+
+        timeEvent_->setEnabled(false);
+        server_->deferredFlush();
+        focusOutWrapper();
     } else {
         // This should not happen, but just in case.
         delete ic;
@@ -142,32 +198,61 @@ void WaylandIMInputContextV1::deactivate(wayland::ZwpInputMethodContextV1 *ic) {
 }
 
 void WaylandIMInputContextV1::repeat() {
-    KeyEvent event(
-        this,
-        Key(repeatSym_, server_->modifiers_ | KeyState::Repeat, repeatKey_ + 8),
-        false, repeatTime_);
-    if (!keyEvent(event)) {
-        ic_->keysym(serial_, repeatTime_, event.rawKey().sym(),
-                    event.isRelease() ? WL_KEYBOARD_KEY_STATE_RELEASED
-                                      : WL_KEYBOARD_KEY_STATE_PRESSED,
-                    event.rawKey().states());
+    if (!ic_ || !realFocus()) {
+        return;
     }
 
-    timeEvent_->setNextInterval(1000000 / repeatRate_);
+    auto *ic = delegatedInputContext();
+    KeyEvent event(
+        ic,
+        Key(repeatSym_, server_->modifiers_ | KeyState::Repeat, repeatKey_ + 8),
+        false, repeatTime_);
+
+    sendKeyToVK(repeatTime_, event.rawKey().code() - 8,
+                WL_KEYBOARD_KEY_STATE_RELEASED);
+    if (!ic->keyEvent(event)) {
+        sendKeyToVK(repeatTime_, event.rawKey().code() - 8,
+                    WL_KEYBOARD_KEY_STATE_PRESSED);
+    }
+
+    uint64_t interval = 1000000 / repeatRate_;
+    timeEvent_->setTime(timeEvent_->time() + interval);
     timeEvent_->setOneShot();
-    server_->display_->flush();
 }
 
 void WaylandIMInputContextV1::surroundingTextCallback(const char *text,
                                                       uint32_t cursor,
                                                       uint32_t anchor) {
-    surroundingText().setText(text, cursor, anchor);
-    updateSurroundingText();
+    std::string str(text);
+    surroundingText().invalidate();
+    do {
+        auto length = utf8::lengthValidated(str);
+        if (length == utf8::INVALID_LENGTH) {
+            break;
+        }
+        if (cursor > str.size() || anchor > str.size()) {
+            break;
+        }
+        size_t cursorByChar =
+            utf8::lengthValidated(str.begin(), str.begin() + cursor);
+        if (cursorByChar == utf8::INVALID_LENGTH) {
+            break;
+        }
+        size_t anchorByChar =
+            utf8::lengthValidated(str.begin(), str.begin() + anchor);
+        if (anchorByChar == utf8::INVALID_LENGTH) {
+            break;
+        }
+        surroundingText().setText(text, cursorByChar, anchorByChar);
+    } while (0);
+    updateSurroundingTextWrapper();
+}
+void WaylandIMInputContextV1::resetCallback() {
+    delegatedInputContext()->reset();
 }
-void WaylandIMInputContextV1::resetCallback() { reset(ResetReason::Client); }
 void WaylandIMInputContextV1::contentTypeCallback(uint32_t hint,
                                                   uint32_t purpose) {
-    CapabilityFlags flags;
+    CapabilityFlags flags = baseFlags;
     if (hint & ZWP_TEXT_INPUT_V1_CONTENT_HINT_PASSWORD) {
         flags |= CapabilityFlag::Password;
     }
@@ -238,12 +323,36 @@ void WaylandIMInputContextV1::contentTypeCallback(uint32_t hint,
     if (purpose == ZWP_TEXT_INPUT_V1_CONTENT_PURPOSE_TERMINAL) {
         flags |= CapabilityFlag::Terminal;
     }
-    setCapabilityFlags(flags);
+    setCapabilityFlagsWrapper(flags);
 }
 void WaylandIMInputContextV1::invokeActionCallback(uint32_t button,
                                                    uint32_t index) {
-    FCITX_UNUSED(button);
-    FCITX_UNUSED(index);
+    InvokeActionEvent::Action action;
+    switch (button) {
+    case BTN_LEFT:
+        action = InvokeActionEvent::Action::LeftClick;
+        break;
+    case BTN_RIGHT:
+        action = InvokeActionEvent::Action::RightClick;
+        break;
+    default:
+        return;
+    }
+    auto *ic = delegatedInputContext();
+    auto preedit = ic->inputPanel().clientPreedit().toString();
+    std::string::size_type offset = index;
+    offset = std::min(offset, preedit.length());
+    auto length =
+        utf8::lengthValidated(std::string_view(preedit).substr(0, offset));
+    if (length == utf8::INVALID_LENGTH) {
+        return;
+    }
+    InvokeActionEvent event(action, length, ic);
+
+    if (!realFocus()) {
+        focusInWrapper();
+    }
+    ic->invokeAction(event);
 }
 void WaylandIMInputContextV1::commitStateCallback(uint32_t serial) {
     serial_ = serial;
@@ -270,7 +379,7 @@ void WaylandIMInputContextV1::keymapCallback(uint32_t format, int32_t fd,
         server_->keymap_.reset();
     }
 
-    auto *mapStr = mmap(nullptr, size, PROT_READ, MAP_SHARED, fd, 0);
+    auto *mapStr = mmap(nullptr, size, PROT_READ, MAP_PRIVATE, fd, 0);
     if (mapStr == MAP_FAILED) {
         close(fd);
         return;
@@ -315,6 +424,8 @@ void WaylandIMInputContextV1::keymapCallback(uint32_t format, int32_t fd,
         1 << xkb_keymap_mod_get_index(server_->keymap_.get(), "Hyper");
     server_->stateMask_.meta_mask =
         1 << xkb_keymap_mod_get_index(server_->keymap_.get(), "Meta");
+
+    server_->parent_->wayland()->call<IWaylandModule::reloadXkbOption>();
 }
 
 void WaylandIMInputContextV1::keyCallback(uint32_t serial, uint32_t time,
@@ -323,11 +434,15 @@ void WaylandIMInputContextV1::keyCallback(uint32_t serial, uint32_t time,
     if (!server_->state_) {
         return;
     }
+    if (!ic_) {
+        return;
+    }
 
     // EVDEV OFFSET
     uint32_t code = key + 8;
 
-    KeyEvent event(this,
+    auto *ic = delegatedInputContext();
+    KeyEvent event(ic,
                    Key(static_cast<KeySym>(xkb_state_key_get_one_sym(
                            server_->state_.get(), code)),
                        server_->modifiers_, code),
@@ -341,26 +456,39 @@ void WaylandIMInputContextV1::keyCallback(uint32_t serial, uint32_t time,
             repeatKey_ = key;
             repeatTime_ = time;
             repeatSym_ = event.rawKey().sym();
-            timeEvent_->setNextInterval(repeatDelay_ * 1000);
+            // Let's trick the key event system by fake our first.
+            // Remove 100 from the initial interval.
+            timeEvent_->setNextInterval(repeatDelay_ * 1000 - repeatHackDelay);
             timeEvent_->setOneShot();
         }
     }
 
     WAYLANDIM_DEBUG() << event.key().toString()
                       << " IsRelease=" << event.isRelease();
-    if (!keyEvent(event)) {
-        ic_->keysym(serial, time, event.rawKey().sym(),
-                    event.isRelease() ? WL_KEYBOARD_KEY_STATE_RELEASED
-                                      : WL_KEYBOARD_KEY_STATE_PRESSED,
-                    event.rawKey().states());
+    if (!ic->keyEvent(event)) {
+        ic_->key(serial, time, key, state);
     }
-    server_->display_->flush();
+
+    // This means our engine is being too slow, this is usually transient (e.g.
+    // cold start up due to data loading, high CPU usage etc).
+    // To avoid an undesired repetition, reset the delay the next interval so we
+    // can handle the release first.
+    if (timeEvent_->time() < now(timeEvent_->clock()) &&
+        timeEvent_->isOneShot()) {
+        WAYLANDIM_DEBUG() << "Engine handling speed can not keep up with key "
+                             "repetition rate.";
+        timeEvent_->setNextInterval(
+            std::min(1000, repeatDelay_ * 1000 - repeatHackDelay));
+    }
+
+    server_->deferredFlush();
 }
 void WaylandIMInputContextV1::modifiersCallback(uint32_t serial,
                                                 uint32_t mods_depressed,
                                                 uint32_t mods_latched,
                                                 uint32_t mods_locked,
                                                 uint32_t group) {
+    FCITX_UNUSED(serial);
     if (!server_->state_) {
         return;
     }
@@ -371,9 +499,8 @@ void WaylandIMInputContextV1::modifiersCallback(uint32_t serial,
                           mods_locked, 0, 0, group);
     server_->instance()->updateXkbStateMask(
         server_->group()->display(), mods_depressed, mods_latched, mods_locked);
-    mask = xkb_state_serialize_mods(
-        server_->state_.get(), static_cast<xkb_state_component>(
-                                   XKB_STATE_DEPRESSED | XKB_STATE_LATCHED));
+    mask = xkb_state_serialize_mods(server_->state_.get(),
+                                    XKB_STATE_MODS_EFFECTIVE);
 
     server_->modifiers_ = 0;
     if (mask & server_->stateMask_.shift_mask) {
@@ -388,28 +515,60 @@ void WaylandIMInputContextV1::modifiersCallback(uint32_t serial,
     if (mask & server_->stateMask_.mod1_mask) {
         server_->modifiers_ |= KeyState::Alt;
     }
+    if (mask & server_->stateMask_.mod2_mask) {
+        server_->modifiers_ |= KeyState::NumLock;
+    }
     if (mask & server_->stateMask_.super_mask) {
         server_->modifiers_ |= KeyState::Super;
     }
+    if (mask & server_->stateMask_.mod4_mask) {
+        server_->modifiers_ |= KeyState::Super;
+    }
     if (mask & server_->stateMask_.hyper_mask) {
         server_->modifiers_ |= KeyState::Hyper;
     }
+    if (mask & server_->stateMask_.mod3_mask) {
+        server_->modifiers_ |= KeyState::Hyper;
+    }
+    if (mask & server_->stateMask_.mod5_mask) {
+        server_->modifiers_ |= KeyState::Mod5;
+    }
     if (mask & server_->stateMask_.meta_mask) {
         server_->modifiers_ |= KeyState::Meta;
     }
-
-    ic_->modifiers(serial, mods_depressed, mods_depressed, mods_latched, group);
 }
 
 void WaylandIMInputContextV1::repeatInfoCallback(int32_t rate, int32_t delay) {
     repeatRate_ = rate;
     repeatDelay_ = delay;
-    timeEvent_->setAccuracy(std::min(delay * 1000, 1000000 / rate));
 }
 
-void WaylandIMInputContextV1::updatePreeditImpl() {
+void WaylandIMInputContextV1::sendKey(uint32_t time, uint32_t sym,
+                                      uint32_t state, KeyStates states) const {
+    if (!ic_) {
+        return;
+    }
+    auto modifiers = toModifiers(states);
+    ic_->keysym(serial_, time, sym, state, modifiers);
+    server_->deferredFlush();
+}
+
+void WaylandIMInputContextV1::sendKeyToVK(uint32_t time, uint32_t key,
+                                          uint32_t state) {
+    if (!ic_) {
+        return;
+    }
+    ic_->key(serial_, time, key, state);
+    server_->deferredFlush();
+}
+
+void WaylandIMInputContextV1::updatePreeditDelegate(InputContext *ic) const {
+    if (!ic_) {
+        return;
+    }
+
     auto preedit =
-        server_->instance()->outputFilter(this, inputPanel().clientPreedit());
+        server_->instance()->outputFilter(ic, ic->inputPanel().clientPreedit());
 
     for (int i = 0, e = preedit.size(); i < e; i++) {
         if (!utf8::validate(preedit.stringAt(i))) {
@@ -418,13 +577,44 @@ void WaylandIMInputContextV1::updatePreeditImpl() {
     }
 
     ic_->preeditCursor(preedit.cursor());
-    ic_->preeditString(serial_, preedit.toString().c_str(),
-                       preedit.toStringForCommit().c_str());
     unsigned int index = 0;
     for (int i = 0, e = preedit.size(); i < e; i++) {
-        ic_->preeditStyling(index, preedit.stringAt(i).size(),
-                            waylandFormat(preedit.formatAt(i)));
-        index += preedit.stringAt(i).size();
+        if (!preedit.stringAt(i).empty()) {
+            ic_->preeditStyling(index, preedit.stringAt(i).size(),
+                                waylandFormat(preedit.formatAt(i)));
+            index += preedit.stringAt(i).size();
+        }
     }
+    ic_->preeditString(serial_, preedit.toString().c_str(),
+                       preedit.toStringForCommit().c_str());
+    server_->deferredFlush();
+}
+
+void WaylandIMInputContextV1::deleteSurroundingTextDelegate(
+    InputContext *ic, int offset, unsigned int size) const {
+    if (!ic_) {
+        return;
+    }
+
+    size_t cursor = ic->surroundingText().cursor();
+    if (static_cast<ssize_t>(cursor) + offset < 0) {
+        return;
+    }
+
+    const auto &text = ic->surroundingText().text();
+    auto len = utf8::length(text);
+
+    size_t start = cursor + offset;
+    size_t end = cursor + offset + size;
+    if (cursor > len || start > len || end > len) {
+        return;
+    }
+
+    auto startBytes = utf8::ncharByteLength(text.begin(), start);
+    auto cursorBytes = utf8::ncharByteLength(text.begin(), cursor);
+    auto sizeBytes = utf8::ncharByteLength(text.begin() + startBytes, size);
+    ic_->deleteSurroundingText(startBytes - cursorBytes, sizeBytes);
+    ic_->commitString(serial_, "");
+    server_->deferredFlush();
 }
 } // namespace fcitx
