@@ -8,18 +8,23 @@
 #include <sys/mman.h>
 #include "fcitx-utils/unixfd.h"
 #include "fcitx-utils/utf8.h"
+#include "fcitx/inputcontext.h"
+#include "virtualinputcontext.h"
 #include "wayland-text-input-unstable-v3-client-protocol.h"
 #include "waylandim.h"
 #include "wl_seat.h"
 
 namespace fcitx {
 
+constexpr CapabilityFlags baseFlags{CapabilityFlag::Preedit,
+                                    CapabilityFlag::FormattedPreedit,
+                                    CapabilityFlag::SurroundingText};
+
 WaylandIMServerV2::WaylandIMServerV2(wl_display *display, FocusGroup *group,
                                      const std::string &name,
                                      WaylandIMModule *waylandim)
-    : group_(group), name_(name), parent_(waylandim),
-      inputMethodManagerV2_(nullptr), display_(static_cast<wayland::Display *>(
-                                          wl_display_get_user_data(display))) {
+    : WaylandIMServerBase(display, group, name, waylandim),
+      inputMethodManagerV2_(nullptr) {
     display_->requestGlobals<wayland::ZwpInputMethodManagerV2>();
     display_->requestGlobals<wayland::ZwpVirtualKeyboardManagerV1>();
     display_->requestGlobals<wayland::WlSeat>();
@@ -92,6 +97,7 @@ void WaylandIMServerV2::refreshSeat() {
             inputContextManager(), this, seat,
             virtualKeyboardManagerV1_->createVirtualKeyboard(seat.get()));
         ic->setFocusGroup(group_);
+        ic->setCapabilityFlags(baseFlags);
     }
 }
 
@@ -109,7 +115,7 @@ void WaylandIMServerV2::remove(wayland::WlSeat *id) {
 WaylandIMInputContextV2::WaylandIMInputContextV2(
     InputContextManager &inputContextManager, WaylandIMServerV2 *server,
     std::shared_ptr<wayland::WlSeat> seat, wayland::ZwpVirtualKeyboardV1 *vk)
-    : InputContext(inputContextManager), server_(server),
+    : VirtualInputContextGlue(inputContextManager), server_(server),
       seat_(std::move(seat)),
       ic_(server->inputMethodManagerV2()->getInputMethod(seat_.get())),
       vk_(vk) {
@@ -132,11 +138,33 @@ WaylandIMInputContextV2::WaylandIMInputContextV2(
         if (pendingDeactivate_) {
             pendingDeactivate_ = false;
             keyboardGrab_.reset();
-            server_->display_->sync();
-            focusOut();
+            // This is the only place we update wayland xkb mask, so it is ok to
+            // reset it to 0. This breaks the caps lock or num lock. But we have
+            // no other option until we can listen to the mod change globally.
+            server_->instance()->clearXkbStateMask(server_->group()->display());
+
+            timeEvent_->setEnabled(false);
+            if (realFocus()) {
+                if (vkReady_) {
+                    // If last key to vk is press, send a release.
+                    while (!pressedVKKey_.empty()) {
+                        auto [vkkey, vktime] = *pressedVKKey_.begin();
+                        sendKeyToVK(vktime, vkkey,
+                                    WL_KEYBOARD_KEY_STATE_RELEASED);
+                    }
+                    vk_->modifiers(0, 0, 0, 0);
+                    server_->deferredFlush();
+                }
+                focusOutWrapper();
+            }
         }
         if (pendingActivate_) {
             pendingActivate_ = false;
+            // There can be only one grab. Always release old grab first.
+            // It is possible when switching between two client, there will be
+            // two activate. In that case we will have already one grab. The
+            // second request would fail and cause invalid object.
+            keyboardGrab_.reset();
             keyboardGrab_.reset(ic_->grabKeyboard());
             if (!keyboardGrab_) {
                 WAYLANDIM_DEBUG() << "Failed to grab keyboard";
@@ -162,8 +190,8 @@ WaylandIMInputContextV2::WaylandIMInputContextV2(
                         repeatInfoCallback(rate, delay);
                     });
                 repeatInfoCallback(repeatRate_, repeatDelay_);
-                focusIn();
-                server_->display_->sync();
+                focusInWrapper();
+                server_->deferredFlush();
             }
         }
     });
@@ -181,8 +209,14 @@ WaylandIMInputContextV2::WaylandIMInputContextV2(
             repeat();
             return true;
         });
+    timeEvent_->setAccuracy(1);
     timeEvent_->setEnabled(false);
     created();
+
+    if (auto *appMonitor = server->parent_->appMonitor(server->name_)) {
+        virtualICManager_ = std::make_unique<VirtualInputContextManager>(
+            &inputContextManager, this, appMonitor);
+    }
 }
 
 WaylandIMInputContextV2::~WaylandIMInputContextV2() {
@@ -191,31 +225,58 @@ WaylandIMInputContextV2::~WaylandIMInputContextV2() {
 }
 
 void WaylandIMInputContextV2::repeat() {
+    if (!realFocus()) {
+        return;
+    }
+    auto *ic = delegatedInputContext();
     KeyEvent event(
-        this,
+        ic,
         Key(repeatSym_, server_->modifiers_ | KeyState::Repeat, repeatKey_ + 8),
         false, repeatTime_);
-    if (!keyEvent(event)) {
-        vk_->key(repeatTime_, event.rawKey().code() - 8,
-                 event.isRelease() ? WL_KEYBOARD_KEY_STATE_RELEASED
-                                   : WL_KEYBOARD_KEY_STATE_PRESSED);
+    sendKeyToVK(repeatTime_, event.rawKey().code() - 8,
+                WL_KEYBOARD_KEY_STATE_RELEASED);
+    if (!ic->keyEvent(event)) {
+        sendKeyToVK(repeatTime_, event.rawKey().code() - 8,
+                    WL_KEYBOARD_KEY_STATE_PRESSED);
     }
-
-    timeEvent_->setNextInterval(1000000 / repeatRate_);
+    uint64_t interval = 1000000 / repeatRate_;
+    timeEvent_->setTime(timeEvent_->time() + interval);
     timeEvent_->setOneShot();
-    server_->display_->flush();
 }
 
 void WaylandIMInputContextV2::surroundingTextCallback(const char *text,
                                                       uint32_t cursor,
                                                       uint32_t anchor) {
-    surroundingText().setText(text, cursor, anchor);
-    updateSurroundingText();
+    std::string str(text);
+    surroundingText().invalidate();
+    do {
+        auto length = utf8::lengthValidated(str);
+        if (length == utf8::INVALID_LENGTH) {
+            break;
+        }
+        if (cursor > str.size() || anchor > str.size()) {
+            break;
+        }
+        size_t cursorByChar =
+            utf8::lengthValidated(str.begin(), str.begin() + cursor);
+        if (cursorByChar == utf8::INVALID_LENGTH) {
+            break;
+        }
+        size_t anchorByChar =
+            utf8::lengthValidated(str.begin(), str.begin() + anchor);
+        if (anchorByChar == utf8::INVALID_LENGTH) {
+            break;
+        }
+        surroundingText().setText(text, cursorByChar, anchorByChar);
+    } while (0);
+    updateSurroundingTextWrapper();
+}
+void WaylandIMInputContextV2::resetCallback() {
+    delegatedInputContext()->reset();
 }
-void WaylandIMInputContextV2::resetCallback() { reset(ResetReason::Client); }
 void WaylandIMInputContextV2::contentTypeCallback(uint32_t hint,
                                                   uint32_t purpose) {
-    CapabilityFlags flags;
+    CapabilityFlags flags = baseFlags;
 
     if (hint & ZWP_TEXT_INPUT_V3_CONTENT_HINT_COMPLETION) {
         flags |= CapabilityFlag::WordCompletion;
@@ -288,7 +349,7 @@ void WaylandIMInputContextV2::contentTypeCallback(uint32_t hint,
     if (purpose == ZWP_TEXT_INPUT_V3_CONTENT_PURPOSE_TERMINAL) {
         flags |= CapabilityFlag::Terminal;
     }
-    setCapabilityFlags(flags);
+    setCapabilityFlagsWrapper(flags);
 }
 
 void WaylandIMInputContextV2::keymapCallback(uint32_t format, int32_t fd,
@@ -306,18 +367,22 @@ void WaylandIMInputContextV2::keymapCallback(uint32_t format, int32_t fd,
         return;
     }
 
-    if (server_->keymap_) {
-        server_->keymap_.reset();
-    }
-
-    auto *mapStr = mmap(nullptr, size, PROT_READ, MAP_SHARED, fd, 0);
+    auto *mapStr = mmap(nullptr, size, PROT_READ, MAP_PRIVATE, fd, 0);
     if (mapStr == MAP_FAILED) {
         return;
     }
 
-    server_->keymap_.reset(xkb_keymap_new_from_string(
-        server_->context_.get(), static_cast<const char *>(mapStr),
-        XKB_KEYMAP_FORMAT_TEXT_V1, XKB_KEYMAP_COMPILE_NO_FLAGS));
+    const bool keymapChanged =
+        (size != server_->keymapData_.size() ||
+         memcmp(mapStr, server_->keymapData_.data(), size) != 0);
+    if (keymapChanged) {
+        server_->keymapData_.resize(size);
+        server_->keymapData_.assign(static_cast<const char *>(mapStr),
+                                    static_cast<const char *>(mapStr) + size);
+        server_->keymap_.reset(xkb_keymap_new_from_string(
+            server_->context_.get(), static_cast<const char *>(mapStr),
+            XKB_KEYMAP_FORMAT_TEXT_V1, XKB_KEYMAP_COMPILE_NO_FLAGS));
+    }
 
     munmap(mapStr, size);
 
@@ -354,7 +419,12 @@ void WaylandIMInputContextV2::keymapCallback(uint32_t format, int32_t fd,
     server_->stateMask_.meta_mask =
         1 << xkb_keymap_mod_get_index(server_->keymap_.get(), "Meta");
 
-    vk_->keymap(format, scopeFD.fd(), size);
+    if (keymapChanged) {
+        vk_->keymap(format, scopeFD.fd(), size);
+        vkReady_ = true;
+    }
+
+    server_->parent_->wayland()->call<IWaylandModule::reloadXkbOption>();
 }
 
 void WaylandIMInputContextV2::keyCallback(uint32_t, uint32_t time, uint32_t key,
@@ -364,11 +434,15 @@ void WaylandIMInputContextV2::keyCallback(uint32_t, uint32_t time, uint32_t key,
         return;
     }
 
-    WAYLANDIM_DEBUG() << "RECEIVE KEY";
+    if (!realFocus()) {
+        focusInWrapper();
+    }
+
     // EVDEV OFFSET
     uint32_t code = key + 8;
 
-    KeyEvent event(this,
+    auto *ic = delegatedInputContext();
+    KeyEvent event(ic,
                    Key(static_cast<KeySym>(xkb_state_key_get_one_sym(
                            server_->state_.get(), code)),
                        server_->modifiers_, code),
@@ -382,19 +456,34 @@ void WaylandIMInputContextV2::keyCallback(uint32_t, uint32_t time, uint32_t key,
             repeatKey_ = key;
             repeatTime_ = time;
             repeatSym_ = event.rawKey().sym();
-            timeEvent_->setNextInterval(repeatDelay_ * 1000);
+            // Let's trick the key event system by fake our first.
+            // Remove 100 from the initial interval.
+            timeEvent_->setNextInterval(repeatDelay_ * 1000 - repeatHackDelay);
             timeEvent_->setOneShot();
         }
     }
 
     WAYLANDIM_DEBUG() << event.key().toString()
                       << " IsRelease=" << event.isRelease();
-    if (!keyEvent(event)) {
-        vk_->key(time, event.rawKey().code() - 8,
-                 event.isRelease() ? WL_KEYBOARD_KEY_STATE_RELEASED
-                                   : WL_KEYBOARD_KEY_STATE_PRESSED);
+    if (!ic->keyEvent(event)) {
+        sendKeyToVK(time, event.rawKey().code() - 8,
+                    event.isRelease() ? WL_KEYBOARD_KEY_STATE_RELEASED
+                                      : WL_KEYBOARD_KEY_STATE_PRESSED);
     }
-    server_->display_->flush();
+
+    // This means our engine is being too slow, this is usually transient (e.g.
+    // cold start up due to data loading, high CPU usage etc).
+    // To avoid an undesired repetition, reset the delay the next interval so we
+    // can handle the release first.
+    if (timeEvent_->time() < now(timeEvent_->clock()) &&
+        timeEvent_->isOneShot()) {
+        WAYLANDIM_DEBUG() << "Engine handling speed can not keep up with key "
+                             "repetition rate.";
+        timeEvent_->setNextInterval(
+            std::min(1000, repeatDelay_ * 1000 - repeatHackDelay));
+    }
+
+    server_->deferredFlush();
 }
 void WaylandIMInputContextV2::modifiersCallback(uint32_t,
                                                 uint32_t mods_depressed,
@@ -428,37 +517,153 @@ void WaylandIMInputContextV2::modifiersCallback(uint32_t,
     if (mask & server_->stateMask_.mod1_mask) {
         server_->modifiers_ |= KeyState::Alt;
     }
+    if (mask & server_->stateMask_.mod2_mask) {
+        server_->modifiers_ |= KeyState::NumLock;
+    }
     if (mask & server_->stateMask_.super_mask) {
         server_->modifiers_ |= KeyState::Super;
     }
+    if (mask & server_->stateMask_.mod4_mask) {
+        server_->modifiers_ |= KeyState::Super;
+    }
     if (mask & server_->stateMask_.hyper_mask) {
         server_->modifiers_ |= KeyState::Hyper;
     }
+    if (mask & server_->stateMask_.mod3_mask) {
+        server_->modifiers_ |= KeyState::Hyper;
+    }
+    if (mask & server_->stateMask_.mod5_mask) {
+        server_->modifiers_ |= KeyState::Mod5;
+    }
     if (mask & server_->stateMask_.meta_mask) {
         server_->modifiers_ |= KeyState::Meta;
     }
 
-    vk_->modifiers(mods_depressed, mods_depressed, mods_latched, group);
+    if (vkReady_) {
+        vk_->modifiers(mods_depressed, mods_latched, mods_locked, group);
+    }
 }
 
 void WaylandIMInputContextV2::repeatInfoCallback(int32_t rate, int32_t delay) {
     repeatRate_ = rate;
     repeatDelay_ = delay;
-    timeEvent_->setAccuracy(std::min(delay * 1000, 1000000 / rate));
 }
 
-void WaylandIMInputContextV2::updatePreeditImpl() {
-    auto preedit =
-        server_->instance()->outputFilter(this, inputPanel().clientPreedit());
+void WaylandIMInputContextV2::sendKeyToVK(uint32_t time, uint32_t key,
+                                          uint32_t state) const {
+    if (!vkReady_) {
+        return;
+    }
+    // Erase old to ensure order, and released ones can the be removed.
+    pressedVKKey_.erase(key);
+    if (state == WL_KEYBOARD_KEY_STATE_PRESSED) {
+        pressedVKKey_[key] = time;
+    }
+    vk_->key(time, key, state);
+    server_->deferredFlush();
+}
 
+void WaylandIMInputContextV2::forwardKeyDelegate(
+    InputContext *, const ForwardKeyEvent &key) const {
+    uint32_t code = 0;
+    if (key.rawKey().code()) {
+        code = key.rawKey().code();
+    } else if (auto xkbState = server_->xkbState()) {
+        auto *map = xkb_state_get_keymap(xkbState);
+        auto min = xkb_keymap_min_keycode(map),
+             max = xkb_keymap_max_keycode(map);
+        for (auto keyCode = min; keyCode < max; keyCode++) {
+            if (xkb_state_key_get_one_sym(xkbState, keyCode) ==
+                static_cast<uint32_t>(key.rawKey().sym())) {
+                code = keyCode;
+                break;
+            }
+        }
+    }
+    sendKeyToVK(time_, code - 8,
+                key.isRelease() ? WL_KEYBOARD_KEY_STATE_RELEASED
+                                : WL_KEYBOARD_KEY_STATE_PRESSED);
+    if (!key.isRelease()) {
+        sendKeyToVK(time_, code - 8, WL_KEYBOARD_KEY_STATE_RELEASED);
+    }
+}
+
+void WaylandIMInputContextV2::updatePreeditDelegate(InputContext *ic) const {
+    if (!realFocus()) {
+        return;
+    }
+    auto preedit =
+        server_->instance()->outputFilter(ic, ic->inputPanel().clientPreedit());
+
+    int highlightStart = -1, highlightEnd = -1;
+    int start = 0, end = 0;
+    bool multipleHighlight = false;
     for (int i = 0, e = preedit.size(); i < e; i++) {
         if (!utf8::validate(preedit.stringAt(i))) {
             return;
         }
+        end = start + preedit.stringAt(i).size();
+        if (preedit.formatAt(i).test(TextFormatFlag::HighLight)) {
+            if (highlightStart == -1) {
+                highlightStart = start;
+                highlightEnd = end;
+            } else if (highlightEnd == start) {
+                highlightEnd = end;
+            } else {
+                multipleHighlight = true;
+            }
+        }
+        start = end;
     }
 
-    ic_->setPreeditString(preedit.toString().data(), preedit.cursor(),
-                          preedit.cursor());
+    int cursorStart = preedit.cursor();
+    int cursorEnd = preedit.cursor();
+    if (!multipleHighlight && highlightStart >= 0 && highlightEnd >= 0) {
+        if (cursorStart == highlightStart) {
+            cursorEnd = highlightEnd;
+        }
+    }
+
+    if (preedit.textLength()) {
+        ic_->setPreeditString(preedit.toString().data(), cursorStart,
+                              cursorEnd);
+    }
     ic_->commit(serial_);
+    server_->deferredFlush();
+}
+
+void WaylandIMInputContextV2::deleteSurroundingTextDelegate(
+    InputContext *ic, int offset, unsigned int size) const {
+    if (!realFocus()) {
+        return;
+    }
+
+    // Cant convert to before/after.
+    if (offset > 0 || offset + static_cast<ssize_t>(size) < 0) {
+        return;
+    }
+
+    size_t cursor = ic->surroundingText().cursor();
+    if (static_cast<ssize_t>(cursor) + offset < 0) {
+        return;
+    }
+
+    const auto &text = ic->surroundingText().text();
+    auto len = utf8::length(text);
+
+    size_t start = cursor + offset;
+    size_t end = cursor + offset + size;
+    // validate length.
+    if (cursor > len || start > len || end > len) {
+        return;
+    }
+
+    auto startBytes = utf8::ncharByteLength(text.begin(), start);
+    auto cursorBytes = utf8::ncharByteLength(text.begin(), cursor);
+    auto sizeBytes = utf8::ncharByteLength(text.begin() + startBytes, size);
+    ic_->deleteSurroundingText(cursorBytes - startBytes,
+                               startBytes + sizeBytes - cursorBytes);
+    ic_->commit(serial_);
+    server_->deferredFlush();
 }
 } // namespace fcitx
