@@ -11,11 +11,13 @@
 #include <xcb-imdkit/encoding.h>
 #include <xcb/xcb_aux.h>
 #include <xkbcommon/xkbcommon.h>
+#include "fcitx-utils/misc_p.h"
 #include "fcitx-utils/stringutils.h"
 #include "fcitx-utils/utf8.h"
 #include "fcitx/focusgroup.h"
 #include "fcitx/inputcontext.h"
 #include "fcitx/instance.h"
+#include "fcitx/userinterface.h"
 
 FCITX_DEFINE_LOG_CATEGORY(xim, "xim")
 FCITX_DEFINE_LOG_CATEGORY(xim_key, "xim_key")
@@ -43,10 +45,9 @@ uint32_t onthespot_style_array[] = {
 };
 
 char COMPOUND_TEXT[] = "COMPOUND_TEXT";
+char UTF8_STRING[] = "UTF8_STRING";
 
-char *encoding_array[] = {
-    COMPOUND_TEXT,
-};
+char *encoding_array[] = {COMPOUND_TEXT, UTF8_STRING};
 
 xcb_im_encodings_t encodings = {FCITX_ARRAY_SIZE(encoding_array),
                                 encoding_array};
@@ -102,18 +103,21 @@ public:
             conn, XCB_COPY_FROM_PARENT, serverWindow_, screen->root, 0, 0, 1, 1,
             1, XCB_WINDOW_CLASS_INPUT_OUTPUT, screen->root_visual, 0, nullptr);
 
+        auto inputStyles =
+            (*parent_->config().useOnTheSpot ? &onthespot_styles : &styles);
+        for (uint32_t i = 0; i < inputStyles->nStyles; i++) {
+            supportedStyles_.insert(inputStyles->styles[i]);
+        }
         im_.reset(xcb_im_create(
             conn, defaultScreen, serverWindow_, guess_server_name().c_str(),
-            XCB_IM_ALL_LOCALES,
-            (*parent_->config().useOnTheSpot ? &onthespot_styles : &styles),
-            nullptr, nullptr, &encodings,
+            XCB_IM_ALL_LOCALES, inputStyles, nullptr, nullptr, &encodings,
             XCB_EVENT_MASK_KEY_PRESS | XCB_EVENT_MASK_KEY_RELEASE,
             &XIMServer::callback, this));
 
         if (::xim().checkLogLevel(LogLevel::Debug)) {
             xcb_im_set_log_handler(im_.get(), XimLogFunc);
         }
-        xcb_im_set_use_sync_mode(im_.get(), false);
+        xcb_im_set_use_sync_mode(im_.get(), true);
 
         filter_ = parent_->xcb()->call<fcitx::IXCBModule::addEventFilter>(
             name, [this](xcb_connection_t *, xcb_generic_event_t *event) {
@@ -139,6 +143,7 @@ public:
     }
 
     Instance *instance() { return parent_->instance(); }
+    const auto &supportedStyles() { return supportedStyles_; }
 
     ~XIMServer() {
         if (im_) {
@@ -176,6 +181,9 @@ private:
     xcb_window_t serverWindow_;
     xcb_ewmh_connection_t *ewmh_;
     std::unique_ptr<HandlerTableEntry<XCBEventFilter>> filter_;
+    // bool value: isUtf8
+    std::unordered_map<xcb_im_client_t *, bool> clientEncodingMapping_;
+    std::unordered_set<uint32_t> supportedStyles_;
 };
 
 pid_t getWindowPid(xcb_ewmh_connection_t *ewmh, xcb_window_t w) {
@@ -217,15 +225,15 @@ std::string getProgramName(XIMServer *server, xcb_im_input_context_t *ic) {
 class XIMInputContext final : public InputContext {
 public:
     XIMInputContext(InputContextManager &inputContextManager, XIMServer *server,
-                    xcb_im_input_context_t *ic)
+                    xcb_im_input_context_t *ic, bool useUtf8)
         : InputContext(inputContextManager, getProgramName(server, ic)),
-          server_(server), xic_(ic) {
+          server_(server), xic_(ic), useUtf8_(useUtf8) {
         setFocusGroup(server->focusGroup());
         xcb_im_input_context_set_data(xic_, this, nullptr);
-        auto style = xcb_im_input_context_get_input_style(ic);
+
         created();
         CapabilityFlags flags = CapabilityFlag::ReportKeyRepeat;
-        if (style & XCB_IM_PreeditCallbacks) {
+        if (validatedInputStyle() & XCB_IM_PreeditCallbacks) {
             flags = flags | CapabilityFlag::Preedit;
             flags = flags | CapabilityFlag::FormattedPreedit;
         }
@@ -236,8 +244,33 @@ public:
         destroy();
     }
 
+    uint32_t validatedInputStyle() {
+        auto style = xcb_im_input_context_get_input_style(xic_);
+        if (server_->supportedStyles().count(style)) {
+            return style;
+        }
+        auto preeditStyle = (style & 0xff) | XCB_IM_StatusNothing;
+        if (server_->supportedStyles().count(preeditStyle)) {
+            return preeditStyle;
+        }
+        auto statusStyle = (style & 0xff00) | XCB_IM_PreeditNothing;
+        if (server_->supportedStyles().count(statusStyle)) {
+            return statusStyle;
+        }
+
+        return XCB_IM_StatusNothing | XCB_IM_PreeditNothing;
+    }
+
     const char *frontend() const override { return "xim"; }
 
+    void maybeUpdateCursorLocationForRootStyle() {
+        if ((validatedInputStyle() & XCB_IM_PreeditPosition) ==
+            XCB_IM_PreeditPosition) {
+            return;
+        }
+        updateCursorLocation();
+    }
+
     void updateCursorLocation() {
         // kinds of like notification for position moving
         bool hasSpotLocation =
@@ -256,11 +289,11 @@ public:
                 server_->conn(), w, server_->root(), p.x, p.y);
             auto reply = makeUniqueCPtr(xcb_translate_coordinates_reply(
                 server_->conn(), trans_cookie, nullptr));
-            if (reply) {
-                setCursorRect(Rect()
-                                  .setPosition(reply->dst_x, reply->dst_y)
-                                  .setSize(0, 0));
+            if (!reply) {
+                return;
             }
+            setCursorRect(
+                Rect().setPosition(reply->dst_x, reply->dst_y).setSize(0, 0));
         } else {
             auto getgeo_cookie = xcb_get_geometry(server_->conn(), w);
             auto reply = makeUniqueCPtr(xcb_get_geometry_reply(
@@ -269,9 +302,12 @@ public:
                 return;
             }
             auto trans_cookie = xcb_translate_coordinates(
-                server_->conn(), w, server_->root(), reply->x, reply->y);
+                server_->conn(), w, server_->root(), 0, 0);
             auto trans_reply = makeUniqueCPtr(xcb_translate_coordinates_reply(
                 server_->conn(), trans_cookie, nullptr));
+            if (!trans_reply) {
+                return;
+            }
 
             setCursorRect(Rect()
                               .setPosition(trans_reply->dst_x,
@@ -322,16 +358,23 @@ public:
 
 protected:
     void commitStringImpl(const std::string &text) override {
-        size_t compoundTextLength;
-        UniqueCPtr<char> compoundText(xcb_utf8_to_compound_text(
-            text.c_str(), text.size(), &compoundTextLength));
-        if (!compoundText) {
-            return;
+        UniqueCPtr<char> compoundText;
+        const char *commit = text.data();
+        size_t length = text.size();
+        if (!useUtf8_) {
+            size_t compoundTextLength;
+            compoundText.reset(xcb_utf8_to_compound_text(
+                text.c_str(), text.size(), &compoundTextLength));
+            if (!compoundText) {
+                return;
+            }
+            commit = compoundText.get();
+            length = compoundTextLength;
         }
         XIM_DEBUG() << "XIM commit: " << text;
 
-        xcb_im_commit_string(server_->im(), xic_, XCB_XIM_LOOKUP_CHARS,
-                             compoundText.get(), compoundTextLength, 0);
+        xcb_im_commit_string(server_->im(), xic_, XCB_XIM_LOOKUP_CHARS, commit,
+                             length, 0);
     }
     void deleteSurroundingTextImpl(int, unsigned int) override {}
     void forwardKeyImpl(const ForwardKeyEvent &key) override {
@@ -392,7 +435,7 @@ protected:
         }
 
         if (!strPreedit.empty() && !preeditStarted) {
-            xcb_im_preedit_start(server_->im(), xic_);
+            xcb_im_preedit_start_callback(server_->im(), xic_);
             preeditStarted = true;
         }
         if (!strPreedit.empty()) {
@@ -402,7 +445,7 @@ protected:
             }
             feedbackBuffer_.clear();
 
-            for (size_t i = 0, offset = 0; i < text.size(); i++) {
+            for (size_t i = 0; i < text.size(); i++) {
                 auto format = text.formatAt(i);
                 const auto &str = text.stringAt(i);
                 uint32_t feedback = 0;
@@ -415,7 +458,6 @@ protected:
                 unsigned int strLen = utf8::length(str);
                 for (size_t j = 0; j < strLen; j++) {
                     feedbackBuffer_.push_back(feedback);
-                    offset++;
                 }
             }
             feedbackBuffer_.push_back(0);
@@ -428,17 +470,25 @@ protected:
                     utf8::length(strPreedit.begin(),
                                  std::next(strPreedit.begin(), text.cursor()));
             }
+            UniqueCPtr<char> compoundText;
             frame.chg_first = 0;
             frame.chg_length = lastPreeditLength_;
-            size_t compoundTextLength;
-            UniqueCPtr<char> compoundText(xcb_utf8_to_compound_text(
-                strPreedit.c_str(), strPreedit.size(), &compoundTextLength));
-            if (!compoundText) {
-                return;
+            if (useUtf8_) {
+                frame.preedit_string =
+                    reinterpret_cast<uint8_t *>(strPreedit.data());
+                frame.length_of_preedit_string = strPreedit.size();
+            } else {
+                size_t compoundTextLength;
+                compoundText.reset(xcb_utf8_to_compound_text(
+                    strPreedit.c_str(), strPreedit.size(),
+                    &compoundTextLength));
+                if (!compoundText) {
+                    return;
+                }
+                frame.length_of_preedit_string = compoundTextLength;
+                frame.preedit_string =
+                    reinterpret_cast<uint8_t *>(compoundText.get());
             }
-            frame.length_of_preedit_string = compoundTextLength;
-            frame.preedit_string =
-                reinterpret_cast<uint8_t *>(compoundText.get());
             frame.feedback_array.size = feedbackBuffer_.size();
             frame.feedback_array.items = feedbackBuffer_.data();
             frame.status = frame.feedback_array.size ? 0 : 2;
@@ -450,6 +500,7 @@ protected:
 private:
     XIMServer *server_;
     xcb_im_input_context_t *xic_;
+    const bool useUtf8_ = false;
     bool preeditStarted = false;
     int lastPreeditLength_ = 0;
     std::vector<uint32_t> feedbackBuffer_;
@@ -461,10 +512,25 @@ private:
 void XIMServer::callback(xcb_im_client_t *client, xcb_im_input_context_t *xic,
                          const xcb_im_packet_header_fr_t *hdr, void *frame,
                          void *arg) {
-    FCITX_UNUSED(client);
     FCITX_UNUSED(hdr);
     FCITX_UNUSED(frame);
-    FCITX_UNUSED(arg);
+
+    switch (hdr->major_opcode) {
+    case XCB_XIM_ENCODING_NEGOTIATION:
+        if (arg) {
+            auto encodingIndex = *static_cast<uint16_t *>(arg);
+            XIM_DEBUG() << "Client encoding: " << client << " "
+                        << encodingIndex;
+            if (encodingIndex != 0) {
+                clientEncodingMapping_[client] = encodingIndex == 1;
+            }
+        }
+        return;
+    case XCB_XIM_DISCONNECT:
+        XIM_DEBUG() << "Client disconnect: " << client;
+        clientEncodingMapping_.erase(client);
+        return;
+    }
 
     if (!xic) {
         return;
@@ -481,10 +547,15 @@ void XIMServer::callback(xcb_im_client_t *client, xcb_im_input_context_t *xic,
     }
 
     switch (hdr->major_opcode) {
-    case XCB_XIM_CREATE_IC:
+    case XCB_XIM_CREATE_IC: {
+        bool useUtf8 = false;
+        if (auto entry = findValue(clientEncodingMapping_, client);
+            entry && *entry) {
+            useUtf8 = true;
+        }
         new XIMInputContext(parent_->instance()->inputContextManager(), this,
-                            xic);
-        break;
+                            xic, useUtf8);
+    } break;
     case XCB_XIM_DESTROY_IC:
         delete ic;
         break;
@@ -525,7 +596,7 @@ void XIMServer::callback(xcb_im_client_t *client, xcb_im_input_context_t *xic,
         break;
     }
     case XCB_XIM_RESET_IC:
-        ic->reset(ResetReason::Client);
+        ic->reset();
         break;
     case XCB_XIM_SET_IC_FOCUS:
         ic->focusIn();
@@ -552,6 +623,18 @@ XIMModule::XIMModule(Instance *instance) : instance_(instance) {
         [this](const std::string &name, xcb_connection_t *) {
             servers_.erase(name);
         });
+
+    updateRootStyleCallback_ = instance_->watchEvent(
+        EventType::InputContextFlushUI, EventWatcherPhase::PreInputMethod,
+        [](Event &event) {
+            auto &uiEvent = static_cast<InputContextFlushUIEvent &>(event);
+            auto ic = uiEvent.inputContext();
+            if (uiEvent.component() == UserInterfaceComponent::InputPanel &&
+                ic->frontendName() == "xim") {
+                auto xic = static_cast<XIMInputContext *>(ic);
+                xic->maybeUpdateCursorLocationForRootStyle();
+            }
+        });
 }
 
 XIMModule::~XIMModule() {}
