@@ -333,7 +333,7 @@ fail:
 
 Bus::~Bus() {
     FCITX_D();
-    if (d->attached_) {
+    if (d->loop_) {
         detachEventLoop();
     }
 }
@@ -489,7 +489,7 @@ void DBusDispatchStatusCallback(DBusConnection *, DBusDispatchStatus status,
 
 void Bus::attachEventLoop(EventLoop *loop) {
     FCITX_D();
-    if (d->attached_) {
+    if (d->loop_) {
         return;
     }
     d->loop_ = loop;
@@ -530,7 +530,11 @@ void Bus::detachEventLoop() {
                                                  nullptr, nullptr);
     d->deferEvent_.reset();
     d->loop_ = nullptr;
-    d->attached_ = false;
+}
+
+EventLoop *Bus::eventLoop() const {
+    FCITX_D();
+    return d->loop_;
 }
 
 std::unique_ptr<Slot> Bus::addMatch(const MatchRule &rule,
@@ -553,6 +557,7 @@ std::unique_ptr<Slot> Bus::addMatch(const MatchRule &rule,
 
 std::unique_ptr<Slot> Bus::addFilter(MessageCallback callback) {
     FCITX_D();
+
     auto slot = std::make_unique<DBusFilterSlot>();
     slot->handler_ = d->filterHandlers_.add(std::move(callback));
 
@@ -649,13 +654,21 @@ bool Bus::requestName(const std::string &name, Flags<RequestNameFlag> flags) {
              ? DBUS_NAME_FLAG_ALLOW_REPLACEMENT
              : 0) |
         ((flags & RequestNameFlag::Queue) ? 0 : DBUS_NAME_FLAG_DO_NOT_QUEUE);
-    return dbus_bus_request_name(d->conn_.get(), name.c_str(), d_flags,
-                                 nullptr);
+    auto ret =
+        dbus_bus_request_name(d->conn_.get(), name.c_str(), d_flags, nullptr);
+    if (ret == DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER ||
+        ret == DBUS_REQUEST_NAME_REPLY_ALREADY_OWNER ||
+        ((ret == DBUS_REQUEST_NAME_REPLY_IN_QUEUE ||
+          ret == DBUS_REQUEST_NAME_REPLY_EXISTS) &&
+         (flags & RequestNameFlag::Queue))) {
+        return true;
+    }
+    return false;
 }
 
 bool Bus::releaseName(const std::string &name) {
     FCITX_D();
-    return dbus_bus_release_name(d->conn_.get(), name.c_str(), nullptr);
+    return dbus_bus_release_name(d->conn_.get(), name.c_str(), nullptr) >= 0;
 }
 
 std::string Bus::serviceOwner(const std::string &name, uint64_t usec) {
