@@ -48,10 +48,45 @@ std::unordered_map<KeySym, const char *, EnumHash> makeLookupKeyNameMap() {
         {FcitxKey_Up, NC_("Key name", "Up")},
         {FcitxKey_Right, NC_("Key name", "Right")},
         {FcitxKey_Down, NC_("Key name", "Down")},
-        {FcitxKey_Page_Up, NC_("Key name", "PgUp")},
-        {FcitxKey_Page_Down, NC_("Key name", "PgDown")},
+        {FcitxKey_Page_Up, NC_("Key name", "Page Up")},
+        {FcitxKey_Page_Down, NC_("Key name", "Page Down")},
         {FcitxKey_Caps_Lock, NC_("Key name", "CapsLock")},
         {FcitxKey_Num_Lock, NC_("Key name", "NumLock")},
+        {FcitxKey_KP_Space, NC_("Key name", "Keypad Space")},
+        {FcitxKey_KP_Tab, NC_("Key name", "Keypad Tab")},
+        {FcitxKey_KP_Enter, NC_("Key name", "Keypad Enter")},
+        {FcitxKey_KP_F1, NC_("Key name", "Keypad F1")},
+        {FcitxKey_KP_F2, NC_("Key name", "Keypad F2")},
+        {FcitxKey_KP_F3, NC_("Key name", "Keypad F3")},
+        {FcitxKey_KP_F4, NC_("Key name", "Keypad F4")},
+        {FcitxKey_KP_Home, NC_("Key name", "Keypad Home")},
+        {FcitxKey_KP_Left, NC_("Key name", "Keypad Left")},
+        {FcitxKey_KP_Up, NC_("Key name", "Keypad Up")},
+        {FcitxKey_KP_Right, NC_("Key name", "Keypad Right")},
+        {FcitxKey_KP_Down, NC_("Key name", "Keypad Down")},
+        {FcitxKey_KP_Page_Up, NC_("Key name", "Keypad Page Up")},
+        {FcitxKey_KP_Page_Down, NC_("Key name", "Keypad Page Down")},
+        {FcitxKey_KP_End, NC_("Key name", "Keypad End")},
+        {FcitxKey_KP_Begin, NC_("Key name", "Keypad Begin")},
+        {FcitxKey_KP_Insert, NC_("Key name", "Keypad Insert")},
+        {FcitxKey_KP_Delete, NC_("Key name", "Keypad Delete")},
+        {FcitxKey_KP_Equal, NC_("Key name", "Keypad =")},
+        {FcitxKey_KP_Multiply, NC_("Key name", "Keypad *")},
+        {FcitxKey_KP_Add, NC_("Key name", "Keypad +")},
+        {FcitxKey_KP_Separator, NC_("Key name", "Keypad ,")},
+        {FcitxKey_KP_Subtract, NC_("Key name", "Keypad -")},
+        {FcitxKey_KP_Decimal, NC_("Key name", "Keypad .")},
+        {FcitxKey_KP_Divide, NC_("Key name", "Keypad /")},
+        {FcitxKey_KP_0, NC_("Key name", "Keypad 0")},
+        {FcitxKey_KP_1, NC_("Key name", "Keypad 1")},
+        {FcitxKey_KP_2, NC_("Key name", "Keypad 2")},
+        {FcitxKey_KP_3, NC_("Key name", "Keypad 3")},
+        {FcitxKey_KP_4, NC_("Key name", "Keypad 4")},
+        {FcitxKey_KP_5, NC_("Key name", "Keypad 5")},
+        {FcitxKey_KP_6, NC_("Key name", "Keypad 6")},
+        {FcitxKey_KP_7, NC_("Key name", "Keypad 7")},
+        {FcitxKey_KP_8, NC_("Key name", "Keypad 8")},
+        {FcitxKey_KP_9, NC_("Key name", "Keypad 9")},
         {FcitxKey_Scroll_Lock, NC_("Key name", "ScrollLock")},
         {FcitxKey_Menu, NC_("Key name", "Menu")},
         {FcitxKey_Help, NC_("Key name", "Help")},
@@ -312,7 +347,7 @@ bool Key::isReleaseOfModifier(const Key &key) const {
         keys.emplace_back(FcitxKey_Shift_L, states);
         keys.emplace_back(FcitxKey_Shift_R, states);
     }
-    if (key.states() & KeyState::Super) {
+    if ((key.states() & KeyState::Super) || (key.states() & KeyState::Super2)) {
         keys.emplace_back(FcitxKey_Super_L, states);
         keys.emplace_back(FcitxKey_Super_R, states);
     }
@@ -327,12 +362,19 @@ bool Key::isReleaseOfModifier(const Key &key) const {
 bool Key::check(const Key &key) const {
     auto states = states_ & KeyStates({KeyState::Ctrl_Alt_Shift,
                                        KeyState::Super, KeyState::Mod3});
+    if (states_.test(KeyState::Super2)) {
+        states |= KeyState::Super;
+    }
 
     // key is keycode based, do key code based check.
     if (key.code()) {
         return key.states_ == states && key.code_ == code_;
     }
 
+    if (key.sym() == FcitxKey_None) {
+        return false;
+    }
+
     if (isModifier()) {
         Key keyAlt = *this;
         auto states = states_ & (~keySymToStates(sym_));
@@ -346,7 +388,36 @@ bool Key::check(const Key &key) const {
 }
 
 bool Key::isDigit() const {
-    return !states_ && sym_ >= FcitxKey_0 && sym_ <= FcitxKey_9;
+    return !states_ && ((sym_ >= FcitxKey_0 && sym_ <= FcitxKey_9) ||
+                        (sym_ >= FcitxKey_KP_0 && sym_ <= FcitxKey_KP_9));
+}
+
+int Key::digit() const {
+    if (states_) {
+        return -1;
+    }
+    if (sym_ >= FcitxKey_0 && sym_ <= FcitxKey_9) {
+        return sym_ - FcitxKey_0;
+    }
+    if (sym_ >= FcitxKey_KP_0 && sym_ <= FcitxKey_KP_9) {
+        return sym_ - FcitxKey_KP_0;
+    }
+    return -1;
+}
+
+int Key::digitSelection(KeyStates states) const {
+    auto filteredStates =
+        states_ &
+        KeyStates({KeyState::Ctrl_Alt_Shift, KeyState::Super, KeyState::Mod3});
+    if (filteredStates != states) {
+        return -1;
+    }
+
+    const int idx = Key(sym_).digit();
+    if (idx >= 0) {
+        return (idx + 9) % 10;
+    }
+    return -1;
 }
 
 bool Key::isUAZ() const {
@@ -376,7 +447,11 @@ bool Key::isCursorMove() const {
     return ((sym_ == FcitxKey_Left || sym_ == FcitxKey_Right ||
              sym_ == FcitxKey_Up || sym_ == FcitxKey_Down ||
              sym_ == FcitxKey_Page_Up || sym_ == FcitxKey_Page_Down ||
-             sym_ == FcitxKey_Home || sym_ == FcitxKey_End) &&
+             sym_ == FcitxKey_Home || sym_ == FcitxKey_End ||
+             sym_ == FcitxKey_KP_Left || sym_ == FcitxKey_KP_Right ||
+             sym_ == FcitxKey_KP_Up || sym_ == FcitxKey_KP_Down ||
+             sym_ == FcitxKey_KP_Page_Up || sym_ == FcitxKey_KP_Page_Down ||
+             sym_ == FcitxKey_KP_Home || sym_ == FcitxKey_KP_End) &&
             (states_ == KeyState::Ctrl || states_ == KeyState::Ctrl_Shift ||
              states_ == KeyState::Shift || states_ == KeyState::NoState));
 }
@@ -390,11 +465,18 @@ bool Key::isKeyPad() const {
 
 bool Key::hasModifier() const { return !!(states_ & KeyState::SimpleMask); }
 
+bool Key::isVirtual() const { return states_.test(KeyState::Virtual); }
+
 Key Key::normalize() const {
     Key key(*this);
     /* key state != 0 */
-    key.states_ = key.states_ & KeyStates({KeyState::Ctrl_Alt_Shift,
-                                           KeyState::Super, KeyState::Mod3});
+    key.states_ =
+        key.states_ & KeyStates({KeyState::Ctrl_Alt_Shift, KeyState::Super,
+                                 KeyState::Mod3, KeyState::Super2});
+    if (key.states_.test(KeyState::Super2)) {
+        key.states_ = key.states_.unset(KeyState::Super2);
+        key.states_ |= KeyState::Super;
+    }
     if (key.states_) {
         if (key.states_ != KeyState::Shift && Key(key.sym_).isLAZ()) {
             key.sym_ = static_cast<KeySym>(key.sym_ + FcitxKey_A - FcitxKey_a);
@@ -461,22 +543,26 @@ std::string Key::toString(KeyStringFormat format) const {
     }
 
 #define _APPEND_MODIFIER_STRING(STR, VALUE)                                    \
-    if (states & KeyState::VALUE) {                                            \
+    if (states & VALUE) {                                                      \
         str += STR;                                                            \
         str += "+";                                                            \
     }
     if (format == KeyStringFormat::Portable) {
-        _APPEND_MODIFIER_STRING("Control", Ctrl)
-        _APPEND_MODIFIER_STRING("Alt", Alt)
-        _APPEND_MODIFIER_STRING("Shift", Shift)
-        _APPEND_MODIFIER_STRING("Super", Super)
-        _APPEND_MODIFIER_STRING("Hyper", Mod3)
+        _APPEND_MODIFIER_STRING("Control", KeyState::Ctrl)
+        _APPEND_MODIFIER_STRING("Alt", KeyState::Alt)
+        _APPEND_MODIFIER_STRING("Shift", KeyState::Shift)
+        _APPEND_MODIFIER_STRING("Super",
+                                (KeyStates{KeyState::Super, KeyState::Super2}))
+        _APPEND_MODIFIER_STRING("Hyper",
+                                (KeyStates{KeyState::Hyper, KeyState::Hyper2}))
     } else {
-        _APPEND_MODIFIER_STRING(C_("Key name", "Control"), Ctrl)
-        _APPEND_MODIFIER_STRING(C_("Key name", "Alt"), Alt)
-        _APPEND_MODIFIER_STRING(C_("Key name", "Shift"), Shift)
-        _APPEND_MODIFIER_STRING(C_("Key name", "Super"), Super)
-        _APPEND_MODIFIER_STRING(C_("Key name", "Hyper"), Mod3)
+        _APPEND_MODIFIER_STRING(C_("Key name", "Control"), KeyState::Ctrl)
+        _APPEND_MODIFIER_STRING(C_("Key name", "Alt"), KeyState::Alt)
+        _APPEND_MODIFIER_STRING(C_("Key name", "Shift"), KeyState::Shift)
+        _APPEND_MODIFIER_STRING(C_("Key name", "Super"),
+                                (KeyStates{KeyState::Super, KeyState::Super2}))
+        _APPEND_MODIFIER_STRING(C_("Key name", "Hyper"),
+                                (KeyStates{KeyState::Hyper, KeyState::Hyper2}))
     }
 
 #undef _APPEND_MODIFIER_STRING
