@@ -30,6 +30,9 @@
 #ifdef ENABLE_X11
 #include "xcb_public.h"
 #endif
+#ifdef WAYLAND_FOUND
+#include "wayland_public.h"
+#endif
 
 #define FCITX_DBUS_SERVICE "org.fcitx.Fcitx5"
 #define FCITX_CONTROLLER_DBUS_INTERFACE "org.fcitx.Fcitx.Controller1"
@@ -141,12 +144,8 @@ public:
             });
     }
     void configure() { instance_->configure(); }
-    void configureAddon(const std::string &addon) {
-        instance_->configureAddon(addon);
-    }
-    void configureInputMethod(const std::string &imName) {
-        instance_->configureInputMethod(imName);
-    }
+    void configureAddon(const std::string &) { instance_->configure(); }
+    void configureInputMethod(const std::string &) { instance_->configure(); }
     std::string currentUI() { return instance_->currentUI(); }
     std::string addonForInputMethod(const std::string &imName) {
         return instance_->addonForInputMethod(imName);
@@ -183,6 +182,46 @@ public:
         return {"", {}};
     }
 
+    std::tuple<
+        std::string, std::string, std::string, DBusVariantMap,
+        std::vector<DBusStruct<std::string, std::string, std::string,
+                               std::string, std::string, std::string,
+                               std::string, bool, std::string, DBusVariantMap>>>
+    fullInputMethodGroupInfo(const std::string &inputMethodGroupName) {
+        std::vector<DBusStruct<std::string, std::string, std::string,
+                               std::string, std::string, std::string,
+                               std::string, bool, std::string, DBusVariantMap>>
+            inputMethodEntries;
+
+        const auto &inputMethodManager = instance_->inputMethodManager();
+        const auto &groupName = !inputMethodGroupName.empty()
+                                    ? inputMethodGroupName
+                                    : inputMethodManager.currentGroup().name();
+        const auto *group = inputMethodManager.group(groupName);
+        if (group == nullptr) {
+            return {"", "", "", {}, {}};
+        }
+
+        for (const auto &item : group->inputMethodList()) {
+            const auto *entry = inputMethodManager.entry(item.name());
+            if (entry == nullptr) {
+                continue;
+            }
+
+            inputMethodEntries.emplace_back(std::forward_as_tuple(
+                entry->uniqueName(), entry->name(), entry->nativeName(),
+                entry->icon(), entry->label(), entry->languageCode(),
+                entry->addon(), entry->isConfigurable(), item.layout(),
+                DBusVariantMap()));
+        }
+
+        return {groupName,
+                group->defaultInputMethod(),
+                group->defaultLayout(),
+                {},
+                inputMethodEntries};
+    }
+
     std::vector<DBusStruct<std::string, std::string, std::string, std::string,
                            std::string, std::string, bool>>
     availableInputMethods() {
@@ -525,6 +564,41 @@ public:
                                     "XCB addon is not available.");
     }
 
+    void openWaylandConnection(const std::string &name) {
+#ifdef WAYLAND_FOUND
+        if (auto *wayland = module_->wayland()) {
+            if (!wayland->call<IWaylandModule::openConnection>(name)) {
+                throw dbus::MethodCallError(
+                    "org.freedesktop.DBus.Error.InvalidArgs",
+                    "Failed to create wayland connection.");
+            }
+            return;
+        }
+#else
+        FCITX_UNUSED(name);
+#endif
+        throw dbus::MethodCallError("org.freedesktop.DBus.Error.InvalidArgs",
+                                    "Wayland addon is not available.");
+    }
+
+    void openWaylandConnectionSocket(UnixFD &fd) {
+#ifdef WAYLAND_FOUND
+        if (auto *wayland = module_->wayland()) {
+            if (!wayland->call<IWaylandModule::openConnectionSocket>(
+                    fd.release())) {
+                throw dbus::MethodCallError(
+                    "org.freedesktop.DBus.Error.InvalidArgs",
+                    "Failed to create wayland connection.");
+            }
+            return;
+        }
+#else
+        FCITX_UNUSED(fd);
+#endif
+        throw dbus::MethodCallError("org.freedesktop.DBus.Error.InvalidArgs",
+                                    "Wayland addon is not available.");
+    }
+
     std::string debugInfo() {
         std::stringstream ss;
         instance_->inputContextManager().foreachGroup([&ss](FocusGroup *group) {
@@ -536,8 +610,9 @@ public:
                     ss << fmt::format("{:02x}", static_cast<int>(v));
                 }
                 ss << "] program:" << ic->program()
-                   << " frontend:" << ic->frontend()
-                   << " cap:" << fmt::format("{:x}", ic->capabilityFlags())
+                   << " frontend:" << ic->frontendName() << " cap:"
+                   << fmt::format("{:x}",
+                                  static_cast<uint64_t>(ic->capabilityFlags()))
                    << " focus:" << ic->hasFocus() << std::endl;
                 return true;
             });
@@ -548,18 +623,34 @@ public:
             if (ic->focusGroup()) {
                 return true;
             }
+            if (ic->frontendName() == "dummy") {
+                return true;
+            }
             ss << "  IC [";
             for (auto v : ic->uuid()) {
                 ss << fmt::format("{:02x}", static_cast<int>(v));
             }
             ss << "] program:" << ic->program()
-               << " frontend:" << ic->frontend() << " focus:" << ic->hasFocus()
-               << std::endl;
+               << " frontend:" << ic->frontendName()
+               << " focus:" << ic->hasFocus() << std::endl;
             return true;
         });
         return ss.str();
     }
 
+    void refresh() {
+        deferEvent_ =
+            instance_->eventLoop().addDeferEvent([this](EventSource *) {
+                instance_->refresh();
+                deferEvent_.reset();
+                return false;
+            });
+    }
+
+    bool checkUpdate() { return instance_->checkUpdate(); }
+
+    void save() { return instance_->save(); }
+
 private:
     DBusModule *module_;
     Instance *instance_;
@@ -582,6 +673,9 @@ private:
                                "s", "");
     FCITX_OBJECT_VTABLE_METHOD(currentInputMethodGroup,
                                "CurrentInputMethodGroup", "", "s");
+    FCITX_OBJECT_VTABLE_METHOD(fullInputMethodGroupInfo,
+                               "FullInputMethodGroupInfo", "s",
+                               "sssa{sv}a(sssssssbsa{sv})");
     FCITX_OBJECT_VTABLE_METHOD(availableInputMethods, "AvailableInputMethods",
                                "", "a(ssssssb)");
     FCITX_OBJECT_VTABLE_METHOD(inputMethodGroupInfo, "InputMethodGroupInfo",
@@ -614,7 +708,14 @@ private:
                                "a(sssibbbasas)");
     FCITX_OBJECT_VTABLE_METHOD(setAddonsState, "SetAddonsState", "a(sb)", "");
     FCITX_OBJECT_VTABLE_METHOD(openX11Connection, "OpenX11Connection", "s", "");
+    FCITX_OBJECT_VTABLE_METHOD(openWaylandConnection, "OpenWaylandConnection",
+                               "s", "");
+    FCITX_OBJECT_VTABLE_METHOD(openWaylandConnectionSocket,
+                               "OpenWaylandConnectionSocket", "h", "");
     FCITX_OBJECT_VTABLE_METHOD(debugInfo, "DebugInfo", "", "s");
+    FCITX_OBJECT_VTABLE_METHOD(refresh, "Refresh", "", "");
+    FCITX_OBJECT_VTABLE_METHOD(checkUpdate, "CheckUpdate", "", "b");
+    FCITX_OBJECT_VTABLE_METHOD(save, "Save", "", "");
 };
 
 DBusModule::DBusModule(Instance *instance)
@@ -626,6 +727,10 @@ DBusModule::DBusModule(Instance *instance)
     if (instance_->willTryReplace()) {
         requestFlag |= RequestNameFlag::ReplaceExisting;
     }
+
+    controller_ = std::make_unique<Controller1>(this, instance);
+    bus_->addObjectVTable("/controller", FCITX_CONTROLLER_DBUS_INTERFACE,
+                          *controller_);
     if (!bus_->requestName(FCITX_DBUS_SERVICE, requestFlag)) {
         instance_->exit();
         throw std::runtime_error("Unable to request dbus name. Is there "
@@ -655,10 +760,6 @@ DBusModule::DBusModule(Instance *instance)
         GNOME_HELPER_NAME,
         [this](const std::string &, const std::string &,
                const std::string &newName) { xkbHelperName_ = newName; });
-
-    controller_ = std::make_unique<Controller1>(this, instance);
-    bus_->addObjectVTable("/controller", FCITX_CONTROLLER_DBUS_INTERFACE,
-                          *controller_);
 }
 
 DBusModule::~DBusModule() {}
@@ -706,6 +807,8 @@ bool DBusModule::lockGroup(int group) {
     return msg.send();
 }
 
+bool DBusModule::hasXkbHelper() const { return !xkbHelperName_.empty(); }
+
 class DBusModuleFactory : public AddonFactory {
     AddonInstance *create(AddonManager *manager) override {
         return new DBusModule(manager->instance());
