@@ -6,22 +6,27 @@
  */
 #include "xcbmenu.h"
 #include <pango/pangocairo.h>
+#include <xcb/xcb.h>
 #include <xcb/xcb_aux.h>
 #include <xcb/xcb_keysyms.h>
 
+#include <optional>
 #include <utility>
 #include <xcb/xcb_icccm.h>
+#include "fcitx-utils/log.h"
 #include "fcitx/inputcontext.h"
 #include "fcitx/userinterfacemanager.h"
 #include "common.h"
+#include "theme.h"
 
 namespace fcitx::classicui {
 
 XCBMenu::XCBMenu(XCBUI *ui, MenuPool *pool, Menu *menu)
     : XCBWindow(ui), pool_(pool), menu_(menu) {
-    auto *fontMap = pango_cairo_font_map_get_default();
-    context_.reset(pango_font_map_create_context(fontMap));
-    ui->fontOption().setupPangoContext(context_.get());
+    fontMap_.reset(pango_cairo_font_map_new());
+    fontMapDefaultDPI_ = pango_cairo_font_map_get_resolution(
+        PANGO_CAIRO_FONT_MAP(fontMap_.get()));
+    context_.reset(pango_font_map_create_context(fontMap_.get()));
     if (auto *ic = ui_->parent()->instance()->mostRecentInputContext()) {
         lastRelevantIc_ = ic->watch();
     }
@@ -76,87 +81,51 @@ bool XCBMenu::filterEvent(xcb_generic_event_t *event) {
             break;
         }
         if (buttonPress->detail != XCB_BUTTON_INDEX_1) {
-            hideParents();
-            hideChilds();
-            xcb_flush(ui_->connection());
+            hideAll();
             return true;
         }
-        for (size_t i = 0; i < items_.size(); i++) {
-            if (items_[i].isSeparator_ ||
-                !items_[i].region_.contains(buttonPress->event_x,
-                                            buttonPress->event_y)) {
-                continue;
-            }
-            if (items_[i].hasSubMenu_) {
-                return true;
-            }
-            // Check if actions is still good.
-            auto actions = menu_->actions();
-            if (i >= actions.size()) {
-                break;
-            }
-            auto *ic = lastRelevantIc();
-            if (!ic) {
-                break;
-            }
-
-            auto id = actions[i]->id();
-            auto icRef = ic->watch();
-            // Why we need to delay the event, because we
-            // want to make ic has focus.
-            activateTimer_ =
-                ui_->parent()->instance()->eventLoop().addTimeEvent(
-                    CLOCK_MONOTONIC, now(CLOCK_MONOTONIC) + 30000, 0,
-                    [this, icRef, id](EventSourceTime *, uint64_t) {
-                        // FCITX_INFO() << "Timer Triggered";
-                        if (auto *ic = icRef.get()) {
-
-                            auto *action = ui_->parent()
-                                               ->instance()
-                                               ->userInterfaceManager()
-                                               .lookupActionById(id);
-                            if (action) {
-                                action->activate(ic);
-                            }
-                        }
-                        activateTimer_.reset();
-                        return true;
-                    });
-            break;
+        if (auto *menu =
+                childByPosition(buttonPress->root_x, buttonPress->root_y)) {
+            menu->handleButtonPress(buttonPress->root_x - menu->x_,
+                                    buttonPress->root_y - menu->y_);
+        } else {
+            hideAll();
+            return true;
         }
-        hideParents();
-        hide();
-        xcb_flush(ui_->connection());
         return true;
     }
     case XCB_MOTION_NOTIFY: {
         auto *motion = reinterpret_cast<xcb_motion_notify_event_t *>(event);
-        if (motion->event == wid_) {
-            for (size_t i = 0; i < items_.size(); i++) {
-                if (!items_[i].isSeparator_ &&
-                    items_[i].region_.contains(motion->event_x,
-                                               motion->event_y)) {
-                    setHoveredIndex(i);
-                    break;
-                }
-            }
-            return true;
+        if (motion->event != wid_) {
+            break;
         }
-        break;
+
+        if (auto *menu = childByPosition(motion->root_x, motion->root_y)) {
+            menu->handleMotionNotify(motion->root_x - menu->x_,
+                                     motion->root_y - menu->y_);
+        }
+        return true;
     }
     case XCB_ENTER_NOTIFY: {
         auto *enter = reinterpret_cast<xcb_enter_notify_event_t *>(event);
-        if (enter->event == wid_) {
-            hasMouse_ = true;
+        if (enter->event != wid_) {
+            break;
+        }
+
+        if (auto *menu = childByPosition(enter->root_x, enter->root_y)) {
+            menu->hasMouse_ = true;
             return true;
         }
         break;
     }
     case XCB_LEAVE_NOTIFY: {
         auto *leave = reinterpret_cast<xcb_leave_notify_event_t *>(event);
-        if (leave->event == wid_) {
-            hasMouse_ = false;
-            setHoveredIndex(-1);
+        if (leave->event != wid_) {
+            break;
+        }
+        if (auto *menu = childByPosition(leave->root_x, leave->root_y)) {
+            menu->hasMouse_ = false;
+            menu->setHoveredIndex(-1);
             return true;
         }
         break;
@@ -172,6 +141,68 @@ bool XCBMenu::filterEvent(xcb_generic_event_t *event) {
     return false;
 }
 
+void XCBMenu::handleButtonPress(int eventX, int eventY) {
+    for (size_t i = 0; i < items_.size(); i++) {
+        if (items_[i].isSeparator_ ||
+            !items_[i].region_.contains(eventX, eventY)) {
+            continue;
+        }
+        if (items_[i].hasSubMenu_) {
+            return;
+        }
+        // Check if actions is still good.
+        auto actions = menu_->actions();
+        if (i >= actions.size()) {
+            break;
+        }
+        auto *ic = lastRelevantIc();
+        if (!ic) {
+            ic = ui_->parent()
+                     ->instance()
+                     ->inputContextManager()
+                     .dummyInputContext();
+        }
+
+        auto id = actions[i]->id();
+        auto icRef = ic->watch();
+        // Why we need to delay the event, because we
+        // want to make ic has focus.
+        activateTimer_ = ui_->parent()->instance()->eventLoop().addTimeEvent(
+            CLOCK_MONOTONIC, now(CLOCK_MONOTONIC) + 30000, 0,
+            [this, that = this->watch(), icRef, id](EventSourceTime *,
+                                                    uint64_t) {
+                if (!that.isValid()) {
+                    return true;
+                }
+                // FCITX_INFO() << "Timer Triggered";
+                if (auto *ic = icRef.get()) {
+                    auto *action = ui_->parent()
+                                       ->instance()
+                                       ->userInterfaceManager()
+                                       .lookupActionById(id);
+                    if (action) {
+                        action->activate(ic);
+                    }
+                }
+                activateTimer_.reset();
+                return true;
+            });
+        break;
+    }
+
+    hideAll();
+}
+
+void XCBMenu::handleMotionNotify(int eventX, int eventY) {
+    for (size_t i = 0; i < items_.size(); i++) {
+        if (!items_[i].isSeparator_ &&
+            items_[i].region_.contains(eventX, eventY)) {
+            setHoveredIndex(i);
+            return;
+        }
+    }
+}
+
 void XCBMenu::hide() {
     if (!visible_) {
         return;
@@ -180,6 +211,9 @@ void XCBMenu::hide() {
     visible_ = false;
     setParent(nullptr);
     xcb_unmap_window(ui_->connection(), wid_);
+    if (ui_->pointerGrabber() == this) {
+        ui_->ungrabPointer();
+    }
 }
 
 void XCBMenu::hideParents() {
@@ -198,6 +232,13 @@ void XCBMenu::hideChilds() {
     }
 }
 
+void XCBMenu::hideAll() {
+    hideParents();
+    hide();
+    hideChilds();
+    xcb_flush(ui_->connection());
+}
+
 bool XCBMenu::childHasMouse() const {
     auto ref = child_;
     while (auto *child = ref.get()) {
@@ -209,6 +250,27 @@ bool XCBMenu::childHasMouse() const {
     return false;
 }
 
+XCBMenu *XCBMenu::childByPosition(int rootX, int rootY) {
+    if (ui_->pointerGrabber() != this) {
+        return this;
+    }
+
+    XCBMenu *result = this;
+    while (auto *child = result->child_.get()) {
+        result = child;
+    }
+    while (result) {
+        Rect rect;
+        rect.setPosition(result->x_, result->y_)
+            .setSize(result->width_, result->height_);
+        if (rect.contains(rootX, rootY)) {
+            break;
+        }
+        result = result->parent_.get();
+    }
+    return result;
+}
+
 void XCBMenu::hideTillMenuHasMouseOrTopLevel() {
     // Go to the innermost child.
     auto *menu = this;
@@ -246,19 +308,19 @@ void XCBMenu::setHoveredIndex(int idx) {
     pool_->setPopupMenuTimer(
         ui_->parent()->instance()->eventLoop().addTimeEvent(
             CLOCK_MONOTONIC, now(CLOCK_MONOTONIC) + 300000, 0,
-            [this](EventSourceTime *, uint64_t) {
+            [this, that = this->watch()](EventSourceTime *, uint64_t) {
+                if (!that.isValid()) {
+                    return true;
+                }
                 do {
                     // FCITX_INFO() << this << " in timer";
                     if (hoveredIndex_ >= 0 && subMenuIndex_ == hoveredIndex_) {
                         // Mouse is on same menu item.
-                        if (auto *child = child_.get()) {
-                            child->hideChilds();
-                            xcb_flush(ui_->connection());
-                        }
                         break;
                     }
 
                     if (hoveredIndex_ >= 0) {
+                        setFocus();
                         // FCITX_INFO() << this << " in timer branch 1";
                         // The current subMenu anyway is not the hovered one.
                         hideChilds();
@@ -277,7 +339,8 @@ void XCBMenu::setHoveredIndex(int idx) {
                             // FCITX_INFO() << this << " in timer show submenu "
                             // << newMenu;
                             newMenu->show(
-                                item.first->region_.translated(x_, y_));
+                                item.first->region_.translated(x_, y_),
+                                ConstrainAdjustment::Slide);
                         }
                     } else {
                         /// FCITX_INFO() << this << " in timer branch 2";
@@ -306,10 +369,28 @@ std::pair<MenuItem *, Action *> XCBMenu::actionAt(size_t index) {
     return {&items_[index], actions[index]};
 }
 
+void XCBMenu::updateDPI(int x, int y) {
+    dpi_ = ui_->dpiByPosition(x, y);
+
+    // Unlike pango cairo context, Cairo font map does not accept negative dpi.
+    // Restore to default value instead.
+    if (dpi_ < 0) {
+        pango_cairo_font_map_set_resolution(
+            PANGO_CAIRO_FONT_MAP(fontMap_.get()), fontMapDefaultDPI_);
+    } else {
+        pango_cairo_font_map_set_resolution(
+            PANGO_CAIRO_FONT_MAP(fontMap_.get()), dpi_);
+    }
+    pango_cairo_context_set_resolution(context_.get(), dpi_);
+}
+
 void XCBMenu::update() {
     auto *ic = lastRelevantIc();
     if (!ic) {
-        return;
+        ic = ui_->parent()
+                 ->instance()
+                 ->inputContextManager()
+                 .dummyInputContext();
     }
 
     // Size hint:
@@ -331,8 +412,8 @@ void XCBMenu::update() {
     auto *fontDesc = pango_font_description_from_string(
         ui_->parent()->config().menuFont->c_str());
     pango_context_set_font_description(context_.get(), fontDesc);
-    pango_cairo_context_set_resolution(context_.get(), dpi_);
     pango_font_description_free(fontDesc);
+    ui_->fontOption().setupPangoContext(context_.get());
 
     const auto &textMargin = *theme.menu->textMargin;
     int i = 0;
@@ -362,6 +443,7 @@ void XCBMenu::update() {
 
         // Calculate size for real items.
         auto text = action->shortText(ic);
+        pango_layout_context_changed(item.layout_.get());
         pango_layout_set_text(item.layout_.get(), text.c_str(), text.size());
         item.textWidth_ = item.textHeight_ = 0;
         pango_layout_get_pixel_size(item.layout_.get(), &item.textWidth_,
@@ -390,7 +472,7 @@ void XCBMenu::update() {
         if (item.isSeparator_) {
             item.layoutX_ = width;
             item.layoutY_ = height;
-            height += separator.height();
+            height += (separator.isImage() ? 2 : separator.height());
             prevIsSeparator = true;
             continue;
         }
@@ -434,16 +516,20 @@ void XCBMenu::update() {
     cairo_t *c = cairo_create(prerender());
 
     cairo_set_operator(c, CAIRO_OPERATOR_SOURCE);
-    theme.paint(c, *theme.menu->background, width, height);
+    theme.paint(c, *theme.menu->background, width, height, /*alpha=*/1.0,
+                /*scale=*/1.0);
     cairo_set_operator(c, CAIRO_OPERATOR_OVER);
     for (const auto &item : items_) {
         if (item.isSeparator_) {
             cairo_save(c);
             cairo_translate(c, item.layoutX_, item.layoutY_);
+            const ThemeImage &separator =
+                theme.loadBackground(*theme.menu->separator);
             theme.paint(c, *theme.menu->separator,
                         width - *theme.menu->contentMargin->marginLeft -
                             *theme.menu->contentMargin->marginRight,
-                        -1);
+                        (separator.isImage() ? 2 : -1), /*alpha=*/1.0,
+                        /*scale=*/1.0);
             cairo_restore(c);
             continue;
         }
@@ -452,29 +538,32 @@ void XCBMenu::update() {
             cairo_save(c);
             cairo_translate(c, item.region_.left(), item.region_.top());
             theme.paint(c, *theme.menu->highlight, item.region_.width(),
-                        item.region_.height());
+                        item.region_.height(),
+                        /*alpha=*/1.0, /*scale=*/1.0);
             cairo_restore(c);
         }
 
         if (item.isChecked_) {
             cairo_save(c);
             cairo_translate(c, item.checkBoxX_, item.checkBoxY_);
-            theme.paint(c, *theme.menu->checkBox, -1, -1);
+            theme.paint(c, *theme.menu->checkBox, -1, -1, /*alpha=*/1.0,
+                        /*scale=*/1.0);
             cairo_restore(c);
         }
 
         if (item.hasSubMenu_) {
             cairo_save(c);
             cairo_translate(c, item.subMenuX_, item.subMenuY_);
-            theme.paint(c, *theme.menu->subMenu, -1, -1);
+            theme.paint(c, *theme.menu->subMenu, -1, -1, /*alpha=*/1.0,
+                        /*scale=*/1.0);
             cairo_restore(c);
         }
 
         cairo_save(c);
         if (item.isHighlight_) {
-            cairoSetSourceColor(c, *theme.menu->highlightTextColor);
+            cairoSetSourceColor(c, theme.menuSelectedItemText());
         } else {
-            cairoSetSourceColor(c, *theme.menu->normalColor);
+            cairoSetSourceColor(c, theme.menuText());
         }
         cairo_translate(c, item.layoutX_, item.layoutY_);
         pango_cairo_show_layout(c, item.layout_.get());
@@ -555,7 +644,7 @@ void XCBMenu::setFocus() {
                         XCB_CURRENT_TIME);
 }
 
-void XCBMenu::show(Rect rect) {
+void XCBMenu::show(Rect rect, ConstrainAdjustment adjustY) {
     // FCITX_INFO() << this << " show() " << hoveredIndex_;
     if (visible_) {
         return;
@@ -565,7 +654,7 @@ void XCBMenu::show(Rect rect) {
     subMenuIndex_ = -1;
     int x = rect.left();
     int y = rect.top();
-    dpi_ = ui_->dpiByPosition(x, y);
+    updateDPI(x, y);
     update();
     const Rect *closestScreen = nullptr;
     int shortestDistance = INT_MAX;
@@ -580,29 +669,27 @@ void XCBMenu::show(Rect rect) {
     x = x + rect.width();
 
     if (closestScreen) {
-        int newX, newY;
 
         if (x + width() > closestScreen->right()) {
-            newX = rect.left() - width();
-        } else {
-            newX = x;
+            x = rect.left() - width();
         }
 
+        switch (adjustY) {
+        case ConstrainAdjustment::Slide:
+            if (y + height() > closestScreen->bottom()) {
+                y = closestScreen->bottom() - height();
+            }
+            break;
+        case ConstrainAdjustment::Flip:
+            if (y + height() > closestScreen->bottom()) {
+                y = rect.top() - height();
+            }
+            break;
+        };
+
         if (y < closestScreen->top()) {
-            newY = closestScreen->top();
-        } else {
-            newY = y;
+            y = closestScreen->top();
         }
-
-        if (newY + height() > closestScreen->bottom()) {
-            if (newY > closestScreen->bottom()) {
-                newY = closestScreen->bottom() - height();
-            } else { /* better position the window */
-                newY = newY - height();
-            }
-        }
-        x = newX;
-        y = newY;
     }
 
     xcb_params_configure_window_t wc;
@@ -616,6 +703,9 @@ void XCBMenu::show(Rect rect) {
 
     xcb_map_window(ui_->connection(), wid_);
     setFocus();
+    if (parent_.isNull()) {
+        ui_->grabPointer(this);
+    }
     xcb_flush(ui_->connection());
     x_ = x;
     y_ = y;
