@@ -35,13 +35,10 @@ XCBTrayWindow::XCBTrayWindow(XCBUI *ui) : XCBWindow(ui, 48, 48) {
     }
     groupAction_.setShortText(_("Group"));
     groupAction_.setMenu(&groupMenu_);
-    inputMethodAction_.setShortText(_("Input Method"));
-    inputMethodAction_.setMenu(&inputMethodMenu_);
     configureAction_.setShortText(_("Configure"));
     restartAction_.setShortText(_("Restart"));
     exitAction_.setShortText(_("Exit"));
     menu_.addAction(&groupAction_);
-    menu_.addAction(&inputMethodAction_);
     menu_.addAction(&separatorActions_[0]);
     menu_.addAction(&configureAction_);
     menu_.addAction(&restartAction_);
@@ -56,7 +53,6 @@ XCBTrayWindow::XCBTrayWindow(XCBUI *ui) : XCBWindow(ui, 48, 48) {
 
     auto &uiManager = ui_->parent()->instance()->userInterfaceManager();
     uiManager.registerAction(&groupAction_);
-    uiManager.registerAction(&inputMethodAction_);
     uiManager.registerAction(&configureAction_);
     uiManager.registerAction(&restartAction_);
     uiManager.registerAction(&exitAction_);
@@ -103,7 +99,8 @@ bool XCBTrayWindow::filterEvent(xcb_generic_event_t *event) {
                 XCBMenu *menu = menuPool_.requestMenu(ui_, &menu_, nullptr);
                 menu->show(Rect()
                                .setPosition(press->root_x, press->root_y)
-                               .setSize(1, 1));
+                               .setSize(1, 1),
+                           ConstrainAdjustment::Flip);
             } else if (press->detail == XCB_BUTTON_INDEX_1) {
                 ui_->parent()->instance()->toggle();
             }
@@ -117,6 +114,7 @@ bool XCBTrayWindow::filterEvent(xcb_generic_event_t *event) {
         if (expose->window == wid_) {
             CLASSICUI_DEBUG() << "Tray recevied expose event";
             update();
+            return true;
         }
         break;
     }
@@ -152,6 +150,7 @@ bool XCBTrayWindow::filterEvent(xcb_generic_event_t *event) {
                    property->window == dockWindow_) {
             isHorizontal_ = trayOrientation();
             resizeTrayWindow();
+            return true;
         }
         break;
     }
@@ -168,8 +167,8 @@ void XCBTrayWindow::initTray() {
     sprintf(trayAtomNameBuf, "_NET_SYSTEM_TRAY_S%d", ui_->defaultScreen());
     size_t i = 0;
     for (const auto *name : atom_names) {
-        atoms_[i] = ui_->parent()->xcb()->call<IXCBModule::atom>(ui_->name(),
-                                                                 name, false);
+        atoms_[i] = ui_->parent()->xcb()->call<IXCBModule::atom>(
+            ui_->displayName(), name, false);
         i++;
     }
 }
@@ -344,15 +343,15 @@ void XCBTrayWindow::paint(cairo_t *c) {
     std::string label;
     const InputMethodEntry *entry = nullptr;
     if (ic) {
+        icon = instance->inputMethodIcon(ic);
         entry = instance->inputMethodEntry(ic);
     }
     if (entry) {
-        icon = entry->icon();
         label = entry->label();
     }
 
     const auto &image = theme.loadImage(
-        icon, label, std::min(height(), width()), ImagePurpose::Tray);
+        icon, label, std::min(height(), width()), ui_->parent());
 
     cairo_save(c);
     cairo_set_operator(c, CAIRO_OPERATOR_SOURCE);
@@ -418,7 +417,7 @@ void XCBTrayWindow::resume() {
     addEventMaskToWindow(ui_->connection(), screen->root,
                          XCB_EVENT_MASK_STRUCTURE_NOTIFY);
     dockCallback_ = ui_->parent()->xcb()->call<IXCBModule::addSelection>(
-        ui_->name(), trayAtomNameBuf,
+        ui_->displayName(), trayAtomNameBuf,
         [this](xcb_atom_t) { refreshDockWindow(); });
     refreshDockWindow();
 }
@@ -433,14 +432,14 @@ void XCBTrayWindow::suspend() {
 
 void XCBTrayWindow::updateMenu() {
     updateGroupMenu();
-    updateInputMethodMenu();
 
     auto &imManager = ui_->parent()->instance()->inputMethodManager();
     if (imManager.groupCount() > 1) {
-        menu_.insertAction(&inputMethodAction_, &groupAction_);
+        menu_.insertAction(&separatorActions_[0], &groupAction_);
     } else {
         menu_.removeAction(&groupAction_);
     }
+    updateInputMethodMenu();
     bool start = false;
     for (auto *action : menu_.actions()) {
         if (action == &separatorActions_[0]) {
@@ -506,8 +505,9 @@ void XCBTrayWindow::updateInputMethodMenu() {
         auto &inputMethodAction = inputMethodActions_.back();
         inputMethodAction.setShortText(entry->name());
         inputMethodAction.connect<SimpleAction::Activated>(
-            [this, imName](InputContext *) {
-                ui_->parent()->instance()->setCurrentInputMethod(imName);
+            [this, imName](InputContext *ic) {
+                ui_->parent()->instance()->setCurrentInputMethod(ic, imName,
+                                                                 false);
             });
         inputMethodAction.setCheckable(true);
         inputMethodAction.setChecked(
@@ -516,7 +516,7 @@ void XCBTrayWindow::updateInputMethodMenu() {
 
         auto &uiManager = ui_->parent()->instance()->userInterfaceManager();
         uiManager.registerAction(&inputMethodAction);
-        inputMethodMenu_.addAction(&inputMethodAction);
+        menu_.insertAction(&separatorActions_[0], &inputMethodAction);
     }
 }
 } // namespace fcitx::classicui
