@@ -15,6 +15,7 @@
 #include "fcitx/action.h"
 #include "fcitx/addonmanager.h"
 #include "fcitx/inputcontext.h"
+#include "fcitx/inputmethodengine.h"
 #include "fcitx/inputmethodentry.h"
 #include "fcitx/inputmethodmanager.h"
 #include "fcitx/instance.h"
@@ -23,16 +24,10 @@
 #include "fcitx/userinterfacemanager.h"
 #include "dbus_public.h"
 
-namespace fcitx {
+#define FCITX_NO_XCB
+#include "xcb_public.h"
 
-bool isKDE() {
-    std::string desktop;
-    auto *desktopEnv = getenv("XDG_CURRENT_DESKTOP");
-    if (desktopEnv) {
-        desktop = desktopEnv;
-    }
-    return desktop == "KDE";
-}
+namespace fcitx {
 
 enum class CursorRectMethod {
     SetSpotRect,
@@ -112,19 +107,21 @@ private:
 };
 
 Kimpanel::Kimpanel(Instance *instance)
-    : instance_(instance),
-      bus_(instance_->addonManager().addon("dbus")->call<IDBusModule::bus>()),
+    : instance_(instance), bus_(dbus()->call<IDBusModule::bus>()),
       watcher_(*bus_) {
+    reloadConfig();
     entry_ = watcher_.watchService(
         "org.kde.impanel", [this](const std::string &, const std::string &,
                                   const std::string &newOwner) {
-            FCITX_INFO() << "Kimpanel new owner" << newOwner;
+            FCITX_INFO() << "Kimpanel new owner: " << newOwner;
             setAvailable(!newOwner.empty());
         });
 }
 
 Kimpanel::~Kimpanel() {}
 
+void Kimpanel::reloadConfig() { readAsIni(config_, "conf/kimpanel.conf"); }
+
 void Kimpanel::suspend() {
     eventHandlers_.clear();
     proxy_.reset();
@@ -133,6 +130,8 @@ void Kimpanel::suspend() {
     hasRelativeV2_ = false;
 }
 
+const Configuration *Kimpanel::getConfig() const { return &config_; }
+
 void Kimpanel::registerAllProperties(InputContext *ic) {
     std::vector<std::string> props;
     if (!ic) {
@@ -144,7 +143,9 @@ void Kimpanel::registerAllProperties(InputContext *ic) {
             props.push_back(actionToStatus(action, ic));
         }
     }
-    props.push_back(inputMethodStatus(ic));
+
+    const auto imStatus = inputMethodStatus(ic);
+    props.push_back(imStatus);
 
     if (ic) {
         for (auto group :
@@ -156,6 +157,8 @@ void Kimpanel::registerAllProperties(InputContext *ic) {
     }
 
     proxy_->registerProperties(props);
+    proxy_->updateProperty(imStatus);
+    proxy_->enable(true);
 
     bus_->flush();
 }
@@ -166,9 +169,10 @@ std::string Kimpanel::actionToStatus(Action *action, InputContext *ic) {
     if (action->menu()) {
         type = "menu";
     }
-    return stringutils::concat(
-        "/Fcitx/", action->name(), ":", action->shortText(ic), ":",
-        iconName(action->icon(ic)), ":", action->longText(ic), ":", type);
+    return stringutils::concat("/Fcitx/", action->name(), ":",
+                               action->shortText(ic), ":",
+                               IconTheme::iconName(action->icon(ic)), ":",
+                               action->longText(ic), ":", type);
 }
 
 void Kimpanel::resume() {
@@ -244,12 +248,39 @@ void Kimpanel::resume() {
             registerAllProperties(icEvent.inputContext());
             updateCurrentInputMethod(icEvent.inputContext());
         }));
+    eventHandlers_.emplace_back(instance_->watchEvent(
+        EventType::FocusGroupFocusChanged, EventWatcherPhase::Default,
+        [this](Event &event) {
+            auto &focusEvent =
+                static_cast<FocusGroupFocusChangedEvent &>(event);
+            if (!focusEvent.newFocus() &&
+                lastInputContext_.get() == focusEvent.oldFocus()) {
+                proxy_->showAux(false);
+                proxy_->showPreedit(false);
+                proxy_->showLookupTable(false);
+                bus_->flush();
+            }
+        }));
 }
 
 void Kimpanel::update(UserInterfaceComponent component,
                       InputContext *inputContext) {
     if (component == UserInterfaceComponent::InputPanel) {
-        updateInputPanel(inputContext);
+        if (classicui() && isKDE() &&
+            (stringutils::startsWith(inputContext->frontendName(), "wayland") ||
+             (xcb() &&
+              stringutils::startsWith(inputContext->display(), "x11:") &&
+              xcb()->call<IXCBModule::isXWayland>(
+                  inputContext->display().substr(4))))) {
+            proxy_->showAux(false);
+            proxy_->showPreedit(false);
+            proxy_->showLookupTable(false);
+            static_cast<UserInterface *>(classicui())
+                ->update(component, inputContext);
+            lastInputContext_ = inputContext->watch();
+        } else {
+            updateInputPanel(inputContext);
+        }
     } else if (component == UserInterfaceComponent::StatusArea) {
         registerAllProperties(inputContext);
     }
@@ -354,27 +385,53 @@ void Kimpanel::updateInputPanel(InputContext *inputContext) {
     bus_->flush();
 }
 
+// This is heuristic, but we guranteed that we don't do crazy things with label.
+std::string extractTextForLabel(const std::string &label) {
+    if (label.empty()) {
+        return "";
+    }
+    auto texts = stringutils::split(label, FCITX_WHITESPACE);
+    if (texts.empty()) {
+        return "";
+    }
+
+    return texts[0];
+}
+
 std::string Kimpanel::inputMethodStatus(InputContext *ic) {
-    std::string icon = "input-keyboard";
     std::string label;
     std::string description = _("Not available");
+    std::string altDescription = "";
+    std::string icon = "input-keyboard";
     if (ic) {
-        const auto *entry = instance_->inputMethodEntry(ic);
-        if (entry) {
-            icon = entry->icon();
+        icon = instance_->inputMethodIcon(ic);
+        if (auto entry = instance_->inputMethodEntry(ic)) {
             label = entry->label();
+            if (auto engine = instance_->inputMethodEngine(ic)) {
+                auto subModeLabel = engine->subModeLabel(*entry, *ic);
+                if (!subModeLabel.empty()) {
+                    label = subModeLabel;
+                }
+                altDescription = engine->subMode(*entry, *ic);
+            }
             description = entry->name();
         }
     }
 
+    label = extractTextForLabel(label);
+
     static const bool preferSymbolic = !isKDE();
-    if (preferSymbolic && icon == "input-keyboard") {
+    if (*config_.preferTextIcon) {
+        icon = "";
+        altDescription = description;
+        description = label;
+    } else if (preferSymbolic && icon == "input-keyboard") {
         icon = "input-keyboard-symbolic";
     }
 
-    return stringutils::concat(
-        "/Fcitx/im:", label.empty() ? description : label, ":", iconName(icon),
-        ":", label.empty() ? "" : description, ":menu");
+    return stringutils::concat("/Fcitx/im:", description, ":",
+                               IconTheme::iconName(icon), ":", altDescription,
+                               ":menu,label=", label);
 }
 
 void Kimpanel::updateCurrentInputMethod(InputContext *ic) {
@@ -389,7 +446,7 @@ void Kimpanel::msgV1Handler(dbus::Message &msg) {
     if (msg.member() == "Exit") {
         instance_->exit();
     } else if (msg.member() == "ReloadConfig") {
-        instance_->reloadConfig();
+        instance_->restart();
     } else if (msg.member() == "Restart") {
         instance_->restart();
     } else if (msg.member() == "Configure") {
@@ -408,7 +465,7 @@ void Kimpanel::msgV1Handler(dbus::Message &msg) {
                 }
                 menuitems.push_back(stringutils::concat(
                     "/Fcitx/im/", entry->uniqueName(), ":", entry->name(), ":",
-                    iconName(entry->icon()), "::"));
+                    IconTheme::iconName(entry->icon()), "::"));
             }
             proxy_->execMenu(menuitems);
         } else if (stringutils::startsWith(property, "/Fcitx/im/")) {
@@ -434,6 +491,9 @@ void Kimpanel::msgV1Handler(dbus::Message &msg) {
             if (auto *menu = action->menu()) {
                 std::vector<std::string> menuitems;
                 for (auto *menuAction : menu->actions()) {
+                    if (menuAction->isSeparator()) {
+                        continue;
+                    }
                     menuitems.push_back(actionToStatus(menuAction, ic));
                 }
                 proxy_->execMenu(menuitems);
